# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Any, Dict, List

import torch

from ..utils.record import MetricMeter


class BaseMetric:
    def __init__(self, metric_name, num_classes, classwise_analysis, **kwargs):
        self.metric_name = metric_name
        self.num_classes = num_classes
        self.classwise_analysis = classwise_analysis
        if self.classwise_analysis:
            self.classwise_metric_meters = [MetricMeter(f'{metric_name}_{i}', ':6.2f') for i in range(num_classes)]
        self.metric_meter = MetricMeter(metric_name, ':6.2f')

    def calibrate(self, pred, target, **kwargs):
        raise NotImplementedError


class MetricFactory:
    def __init__(self, task, metrics, metric_adaptor, classwise_analysis) -> None:
        self.task = task
        self.metrics = metrics
        self.metric_adaptor = metric_adaptor
        self.classwise_analysis = classwise_analysis

    def reset_values(self):
        for phase in self.metrics:
            [metric.metric_meter.reset() for metric in self.metrics[phase]]

    def update(self, pred: torch.Tensor, target: torch.Tensor, phase: str, **kwargs: Any) -> None:
        if len(pred) == 0: # Removed dummy batch has 0 len
            return
        kwargs.update(self.metric_adaptor(pred, target))
        for metric in self.metrics[phase.lower()]:
            metric.calibrate(pred, target, **kwargs)

    def result(self, phase='train'):
        ret = {metric.metric_name: {} for metric in self.metrics[phase]} # Initialize with empty dict

        if phase == 'valid' and self.classwise_analysis: # Add classwise results only for valid phase
            for metric in self.metrics[phase]:
                classwise_result_dict = {i:classwise_meter.avg for i, classwise_meter in enumerate(metric.classwise_metric_meters)}
                ret[metric.metric_name] = {'classwise': classwise_result_dict}

        for metric in self.metrics[phase]:
            ret[metric.metric_name]['mean'] = metric.metric_meter.avg # Add mean score

        return ret

    @property
    def metric_names(self):
        return [metric.metric_name for metric in self.metrics[list(self.metrics.keys())[0]]]

    @property
    def primary_metric(self):
        return self.metric_names[0]



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import List 
from omegaconf import DictConfig
import torch 
import torch.nn as nn 
import math

from ....op.custom import SeparableConvLayer
from ....utils import ModelOutput 
from .detection import AnchorGenerator


class YOLOFastestHeadV2(nn.Module): 
    def __init__(
        self,
        num_classes: int, 
        intermediate_features_dim: List[int], 
        params: DictConfig) -> None:
        super().__init__()
        anchors = params.anchors
        self.num_anchors = len(anchors[0]) // 2
        hidden_dim = int(intermediate_features_dim[0])
        self.cls_head = YOLOFastestClassificationHead(hidden_dim, self.num_anchors, num_classes)  
        self.reg_head = YOLOFastestRegressionHead(hidden_dim, self.num_anchors) 

    def forward(self, x):
        cls_logits, objs = self.cls_head(x)
        bbox_regression = self.reg_head(x)
        outputs = list()
        for reg, obj, logits in zip(bbox_regression, objs, cls_logits):
            reg = reg.view(reg.shape[0], self.num_anchors, -1, reg.shape[-2], reg.shape[-1])
            obj = obj.view(obj.shape[0], self.num_anchors, -1, obj.shape[-2], obj.shape[-1])
            logits = logits.repeat(1, self.num_anchors, 1, 1).view(logits.shape[0], self.num_anchors, -1, logits.shape[-2], logits.shape[-1])
            output = torch.cat([reg, obj, logits], 2)
            outputs.append(output)
        return ModelOutput(pred=outputs)

def yolo_fastest_head_v2(num_classes, intermediate_features_dim, conf_model_head) -> YOLOFastestHeadV2:
    return YOLOFastestHeadV2(num_classes=num_classes, 
                             intermediate_features_dim=intermediate_features_dim,
                             params=conf_model_head.params)


class YOLOFastestClassificationHead(nn.Module):
    def __init__(
        self, 
        in_channels,
        num_anchors, 
        num_classes,
        prior_prob = 1e-2,
        num_layers = 2,
        ) -> None:
        super().__init__()
        self.num_classes = num_classes 
        self.num_anchors = num_anchors
        self.layer = nn.ModuleList()
        self.cls_logits = nn.ModuleList()
        self.obj = nn.ModuleList()
        for _ in range(num_layers):
            self.layer.append(
                nn.Sequential(*[
                    SeparableConvLayer(in_channels, in_channels, 5, padding=2, no_out_act=True),
                    SeparableConvLayer(in_channels, in_channels, 5, padding=2, no_out_act=True),
                ])
            )
            self.cls_logits.append(nn.Conv2d(in_channels, num_classes, 1, 1, 0, bias=True))
            self.obj.append(nn.Conv2d(in_channels, num_anchors, 1, 1, 0, bias=True))
        
        self.initialize_biases(prior_prob=prior_prob)
        
    
    def initialize_biases(self, prior_prob):
        for conv in self.cls_logits:
            b = conv.bias.view(1, -1)
            b.data.fill_(-math.log((1 - prior_prob) / prior_prob))
            conv.bias = torch.nn.Parameter(b.view(-1), requires_grad=True)

        for conv in self.obj:
            b = conv.bias.view(1, -1)
            b.data.fill_(-math.log((1 - prior_prob) / prior_prob))
            conv.bias = torch.nn.Parameter(b.view(-1), requires_grad=True)
    
    def forward(self, x): 
        all_cls_logits = [] 
        all_objs = []
        for idx, out in enumerate(x):
            features = self.layer[idx](out)
            cls_logits = features
            objectness = features
            cls_logits = self.cls_logits[idx](cls_logits)
            objectness = self.obj[idx](objectness)
            all_cls_logits.append(cls_logits)
            all_objs.append(objectness)
        return all_cls_logits, all_objs


class YOLOFastestRegressionHead(nn.Module): 
    def __init__(
        self,
        in_channels, 
        num_anchors,
        num_layers=2,
        ) -> None:
        super().__init__()
        self.layer = nn.ModuleList()
        self.bbox_reg = nn.ModuleList()
        for _ in range(num_layers):
            self.layer.append(
                nn.Sequential(*[
                                SeparableConvLayer(in_channels, in_channels, 5, padding=2, no_out_act=True),
                                SeparableConvLayer(in_channels, in_channels, 5, padding=2, no_out_act=True)
                                ]))
            self.bbox_reg.append(nn.Conv2d(in_channels, 4 * num_anchors, 1, 1, 0))
    
    def forward(self, x): 
        all_bbox_regression = []
        for idx, out in enumerate(x):
            features = self.layer[idx](out)
            bbox_regression = features
            bbox_regression = self.bbox_reg[idx](bbox_regression)
            all_bbox_regression.append(bbox_regression)
        return all_bbox_regression




# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import os
from pathlib import Path
from typing import Callable, Dict, List, Type

import torch
import torch.nn as nn
from loguru import logger
from omegaconf import OmegaConf
from torch.nn.parallel import DistributedDataParallel as DDP

from .base import ClassificationModel, DetectionModel, ONNXModel, SegmentationModel, TaskModel, TFLiteModel
from .registry import (
    MODEL_BACKBONE_DICT,
    MODEL_FULL_DICT,
    MODEL_HEAD_DICT,
    MODEL_NECK_DICT,
    SUPPORTING_TASK_LIST,
    TASK_MODEL_DICT,
)
from .utils import get_model_format, load_from_checkpoint


def load_full_model(conf_model, model_name, num_classes, model_checkpoint, use_pretrained):
    model_fn: Callable[..., nn.Module] = MODEL_FULL_DICT[model_name]
    conf_model.architecture.full.nick_name = conf_model.name # @illian01: model.name should be same with conf_model.name
    model: nn.Module = model_fn(num_classes=num_classes, conf_model_full=conf_model.architecture.full)
    if use_pretrained:
        model = load_from_checkpoint(
            model, model_checkpoint,
            load_checkpoint_head=conf_model.checkpoint.load_head,
        )
    model.save_dtype = next(model.parameters()).dtype # If loaded model is float16, save it as float16
    model = model.float() # Train with float32

    return model


def load_backbone_and_head_model(
        conf_model, task, backbone_name, head_name, num_classes,
        model_checkpoint, use_pretrained, freeze_backbone,
    ):
    if task not in TASK_MODEL_DICT:
        raise ValueError(
            f"No such task(s) named: {task}. This should be included in SUPPORTING_TASK_LIST ({SUPPORTING_TASK_LIST})")

    # Backbone construction
    backbone_fn: Callable[..., nn.Module] = MODEL_BACKBONE_DICT[backbone_name]
    backbone: nn.Module = backbone_fn(task=task, conf_model_backbone=conf_model.architecture.backbone)

    # Neck construction
    intermediate_features_dim = backbone.intermediate_features_dim
    neck = None
    if getattr(conf_model.architecture, 'neck', None):
        neck_name = conf_model.architecture.neck.name
        neck_fn: Callable[..., nn.Module] = MODEL_NECK_DICT[neck_name]
        neck = neck_fn(intermediate_features_dim=backbone.intermediate_features_dim, conf_model_neck=conf_model.architecture.neck)
        intermediate_features_dim = neck.intermediate_features_dim

    # Head construction
    head_module = MODEL_HEAD_DICT[task][head_name]
    if task == 'classification':
        head = head_module(num_classes=num_classes, feature_dim=backbone.feature_dim, conf_model_head=conf_model.architecture.head)
    elif task in ['segmentation', 'detection', 'pose_estimation']:
        head = head_module(num_classes=num_classes,
                                intermediate_features_dim=intermediate_features_dim,
                                conf_model_head=conf_model.architecture.head)

    # Assemble model and load checkpoint
    model = TASK_MODEL_DICT[task](conf_model, backbone, neck, head, freeze_backbone)
    if use_pretrained:
        model = load_from_checkpoint(
            model, model_checkpoint,
            load_checkpoint_head=conf_model.checkpoint.load_head,
        )
    model.save_dtype = next(model.parameters()).dtype
    model = model.float() # Train with float32
    return model


def build_model(model_conf, num_classes, devices, distributed) -> nn.Module:

    task = model_conf.task
    model_checkpoint = model_conf.checkpoint.path
    use_pretrained = model_conf.checkpoint.use_pretrained

    model_format = get_model_format(model_conf)

    if model_format == 'torch':
        if model_conf.single_task_model:
            model_name = str(model_conf.architecture.full.name).lower()
            model = load_full_model(
                model_conf, model_name, num_classes,
                model_checkpoint=model_checkpoint,
                use_pretrained=use_pretrained
            )
        else:

            backbone_name = str(model_conf.architecture.backbone.name).lower()
            head_name = str(model_conf.architecture.head.name).lower()
            freeze_backbone = model_conf.freeze_backbone
            model = load_backbone_and_head_model(
                model_conf, task, backbone_name, head_name, num_classes,
                model_checkpoint=model_checkpoint,
                use_pretrained=use_pretrained,
                freeze_backbone=freeze_backbone,
            )

        model = model.to(device=devices)
        if distributed:
            model = DDP(model, device_ids=[devices], find_unused_parameters=True)  # TODO: find_unused_parameters should be false (for now, PIDNet has problem)

        return model

    elif model_format == 'torch.fx':
        assert Path(model_conf.checkpoint.path).exists()
        model = torch.load(model_conf.checkpoint.path)

    elif model_format == 'onnx':
        assert Path(model_conf.checkpoint.path).exists()
        model = ONNXModel(model_conf)
        model.set_provider(devices)

    elif model_format == 'tflite':
        assert Path(model_conf.checkpoint.path).exists()
        model = TFLiteModel(model_conf)

    return model



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import math
from copy import deepcopy

import torch

__all__ = ['ModelEMA', 'ExpDecayModelEMA']


class ModelEMA:
    def __init__(self, model, decay, device=None):
        self.ema_model = deepcopy(model)
        self.ema_model.eval()
        self.decay = decay
        self.device = device
        if self.device is not None:
            self.module.to(device=device)

    def update(self, model):
        with torch.no_grad():
            for ema_param, model_param in zip(self.ema_model.state_dict().values(), model.state_dict().values()):
                if self.device is not None:
                    model_param = model_param.to(device=self.device)
                updated_param = self.decay * ema_param + (1. - self.decay) * model_param
                ema_param.copy_(updated_param)


class ExpDecayModelEMA:
    def __init__(self, model, decay, beta, device=None):
        self.ema_model = deepcopy(model)
        self.ema_model.eval()
        self.decay = decay
        self.beta = beta
        self.device = device
        self.counter = 0
        if self.device is not None:
            self.module.to(device=device)

    def get_decay(self):
        return self.decay * (1 - math.exp(-self.counter / self.beta))

    def update(self, model):
        decay = self.get_decay()
        with torch.no_grad():
            for ema_param, model_param in zip(self.ema_model.state_dict().values(), model.state_dict().values()):
                if self.device is not None:
                    model_param = model_param.to(device=self.device)
                updated_param = decay * ema_param + (1. - decay) * model_param
                ema_param.copy_(updated_param)
        self.counter += 1


def build_ema(model, conf):
    decay_type = conf.training.ema.name
    if decay_type == 'constant_decay':
        return ModelEMA(model=model, decay=conf.training.ema.decay)
    elif decay_type == 'exp_decay':
        return ExpDecayModelEMA(model=model, decay=conf.training.ema.decay, beta=conf.training.ema.beta)
    else:
        raise ValueError(f"Unsupported EMA decay type. training.ema_decay.name: {decay_type}")



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Callable, Dict, Type

from .augmentation.transforms import create_transform
from .base import BaseCustomDataset, BaseHFDataset, BaseSampleLoader
from .classification import (
    ClassficationSampleLoader,
    ClassificationCustomDataset,
    ClassificationHFDataset,
)
from .detection import DetectionCustomDataset, DetectionSampleLoader
from .pose_estimation import (
    PoseEstimationCustomDataset,
    PoseEstimationSampleLoader,
)
from .segmentation import (
    SegmentationCustomDataset,
    SegmentationHFDataset,
    SegmentationSampleLoader,
)

CREATE_TRANSFORM = create_transform

CUSTOM_DATASET: Dict[str, Type[BaseCustomDataset]] = {
    'classification': ClassificationCustomDataset,
    'segmentation': SegmentationCustomDataset,
    'detection': DetectionCustomDataset,
    'pose_estimation': PoseEstimationCustomDataset,
}

HUGGINGFACE_DATASET: Dict[str, Type[BaseHFDataset]] = {
    'classification': ClassificationHFDataset,
    'segmentation': SegmentationHFDataset
}

DATA_SAMPLER: Dict[str, Type[BaseSampleLoader]] = {
    'classification': ClassficationSampleLoader,
    'segmentation': SegmentationSampleLoader,
    'detection': DetectionSampleLoader,
    'pose_estimation': PoseEstimationSampleLoader,
}



import math
from typing import Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch.fx.proxy import Proxy


def generate_anchors(image_shape: Union[int, Tuple[int, int]], strides: List[int]):
    """
        This implementation is based on https://github.com/WongKinYiu/YOLO/blob/main/yolo/utils/bounding_box_utils.py.
    """
    image_height, image_width = image_shape
    anchors = []
    scaler = []

    for stride in strides:
        anchor_num = image_width // stride * image_height // stride
        scaler.append(torch.full((anchor_num,), stride))
        offset = stride // 2
        h = torch.arange(0, image_height, stride) + offset
        w = torch.arange(0, image_width, stride) + offset

        if torch.__version__ >= "2.3.0":
            anchor_h, anchor_w = torch.meshgrid(h, w, indexing="ij")
        else:
            anchor_h, anchor_w = torch.meshgrid(h, w)

        anchor = torch.stack([anchor_w.flatten(), anchor_h.flatten()], dim=-1)
        anchors.append(anchor)
    all_anchors = torch.cat(anchors, dim=0)
    all_scalers = torch.cat(scaler, dim=0)
    return all_anchors, all_scalers

def transform_bbox(bboxes: Union[Tensor, Proxy],
                   indicator="xywh -> xyxy",
                   image_shape: Optional[Union[int, Tuple[int, int]]]=None):
    def is_normalized(fmt: str) -> bool:
        return fmt.endswith('n')

    VALID_IN_TYPE = VALID_OUT_TYPE = ["xyxy", "xyxyn", "xywh", "xywhn", "cxcywh", "cxcywhn"]
    dtype = bboxes.dtype
    in_type, out_type = indicator.replace(" ", "").split("->")
    assert in_type in VALID_IN_TYPE, f"Invalid in_type: '{in_type}'. Must be one of {VALID_IN_TYPE}."
    assert out_type in VALID_OUT_TYPE, f"Invalid out_type: '{out_type}'. Must be one of {VALID_OUT_TYPE}."

    if is_normalized(in_type):
        assert image_shape is not None, f"image_shape is required for normalized conversion: {indicator}"
        if isinstance(image_shape, int):
            img_height = img_width = image_shape
        else:
            img_height, img_width = image_shape
            assert isinstance(img_height, int) and isinstance(img_width, int), \
                f"Invalid type: (height: {type(img_height)}, width: {type(img_width)}. Must be (int, int))"
        in_type = in_type[:-1]
    else:
        img_height = img_width = 1.0

    if in_type == "xyxy":
        x_min, y_min, x_max, y_max = bboxes.unbind(-1)
    elif in_type == "xywh":
        x_min, y_min, w, h = bboxes.unbind(-1)
        x_max = x_min + w
        y_max = y_min + h
    elif in_type == "cxcywh":
        cx, cy, w, h = bboxes.unbind(-1)
        x_min = cx - w / 2
        y_min = cy - h / 2
        x_max = cx + w / 2
        y_max = cy + h / 2

    x_min *= img_width
    y_min *= img_height
    x_max *= img_width
    y_max *= img_height
    assert (x_max >= x_min).all(), "Invalid box: x_max < x_min"
    assert (y_max >= y_min).all(), "Invalid box: y_max < y_min"

    if is_normalized(out_type):
        assert image_shape is not None, f"img_size is required for normalized conversion: {indicator}"
        if isinstance(image_shape, int):
            img_height = img_width = image_shape
        else:
            img_height, img_width = image_shape
            assert isinstance(img_height, int) and isinstance(img_width, int), \
                f"Invalid type: (height: {type(img_height)}, width: {type(img_width)}. Must be (int, int))"
        out_type = out_type[:-1]
    else:
        img_height = img_width = 1.0

    x_min /= img_width
    y_min /= img_height
    x_max /= img_width
    y_max /= img_height
    if out_type == "xywh":
        bbox = torch.stack([x_min, y_min, x_max - x_min, y_max - y_min], dim=-1)
    elif out_type == "xyxy":
        bbox = torch.stack([x_min, y_min, x_max, y_max], dim=-1)
    elif out_type == "cxcywh":
        bbox = torch.stack([(x_min + x_max) / 2, (y_min + y_max) / 2, x_max - x_min, y_max - y_min], dim=-1)

    return bbox.to(dtype=dtype)

def bboxes_iou(bboxes_a, bboxes_b, xyxy=True):
    if bboxes_a.shape[1] != 4 or bboxes_b.shape[1] != 4:
        raise IndexError

    if xyxy:
        tl = torch.max(bboxes_a[:, None, :2], bboxes_b[:, :2])
        br = torch.min(bboxes_a[:, None, 2:], bboxes_b[:, 2:])
        area_a = torch.prod(bboxes_a[:, 2:] - bboxes_a[:, :2], 1)
        area_b = torch.prod(bboxes_b[:, 2:] - bboxes_b[:, :2], 1)
    else:
        tl = torch.max(
            (bboxes_a[:, None, :2] - bboxes_a[:, None, 2:] / 2),
            (bboxes_b[:, :2] - bboxes_b[:, 2:] / 2),
        )
        br = torch.min(
            (bboxes_a[:, None, :2] + bboxes_a[:, None, 2:] / 2),
            (bboxes_b[:, :2] + bboxes_b[:, 2:] / 2),
        )

        area_a = torch.prod(bboxes_a[:, 2:], 1)
        area_b = torch.prod(bboxes_b[:, 2:], 1)
    en = (tl < br).type(tl.type()).prod(dim=2)
    area_i = torch.prod(br - tl, 2) * en  # * ((tl < br).all())
    return area_i / (area_a[:, None] + area_b - area_i)


def calculate_iou(bbox1, bbox2, metric="iou", EPS=1e-7) -> Tensor:
    VALID_METRICS = {"iou", "giou", "diou", "ciou"}
    assert metric.lower() in VALID_METRICS, f"Invalid IoU metric: '{metric}'. Must be one of {VALID_METRICS}"
    metric = metric.lower()

    tl = torch.max(
        (bbox1[:, :2] - bbox1[:, 2:] / 2), (bbox2[:, :2] - bbox2[:, 2:] / 2)
    )
    br = torch.min(
        (bbox1[:, :2] + bbox1[:, 2:] / 2), (bbox2[:, :2] + bbox2[:, 2:] / 2)
    )

    area_p = torch.prod(bbox1[:, 2:], 1)
    area_g = torch.prod(bbox2[:, 2:], 1)

    en = (tl < br).type(tl.type()).prod(dim=1)
    area_i = torch.prod(br - tl, 1) * en
    area_u = area_p + area_g - area_i
    iou = (area_i) / (area_u + EPS)

    if metric == "iou":
        return iou
    elif metric == "giou":
        c_tl = torch.min(
            (bbox1[:, :2] - bbox1[:, 2:] / 2), (bbox2[:, :2] - bbox2[:, 2:] / 2)
        )
        c_br = torch.max(
            (bbox1[:, :2] + bbox1[:, 2:] / 2), (bbox2[:, :2] + bbox2[:, 2:] / 2)
        )
        area_c = torch.prod(c_br - c_tl, 1)
        giou = iou - (area_c - area_u) / area_c.clamp(EPS)
        return giou
    elif metric == "diou" or metric == "ciou":
        cent1 = bbox1[..., :2]  # (cx1, cy1)
        cent2 = bbox2[..., :2]  # (cx2, cy2)

        cent_dist = torch.sum((cent1 - cent2) * (cent1 - cent2), dim=-1)

        c_tl = torch.min(
            bbox1[..., :2] - bbox1[..., 2:] / 2,
            bbox2[..., :2] - bbox2[..., 2:] / 2
        )
        c_br = torch.max(
            bbox1[..., :2] + bbox1[..., 2:] / 2,
            bbox2[..., :2] + bbox2[..., 2:] / 2
        )

        diag_dist = torch.sum((c_br - c_tl) ** 2, dim=-1) + EPS

        diou = iou - (cent_dist / diag_dist)
        if metric == "diou":
            return diou
        arctan = torch.atan(bbox1[..., 2] / (bbox1[..., 3] + EPS)) - torch.atan(bbox2[..., 2] / (bbox2[..., 3] + EPS))
        v = (4 / (math.pi ** 2)) * (arctan ** 2)
        with torch.no_grad():
            alpha = v / (v - iou + 1 + EPS)
        ciou = diou - alpha * v
        return ciou



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Dict, Type

from .classification import ClassificationPostprocessor
from .detection import DetectionPostprocessor
from .pose_estimation import PoseEstimationPostprocessor
from .segmentation import SegmentationPostprocessor

POSTPROCESSOR_DICT = {
    'fc': ClassificationPostprocessor,
    'fc_conv': ClassificationPostprocessor,
    'all_mlp_decoder': SegmentationPostprocessor,
    'anchor_free_decoupled_head': DetectionPostprocessor,
    'pidnet': SegmentationPostprocessor,
    'anchor_decoupled_head': DetectionPostprocessor,
    'yolo_fastest_head_v2': DetectionPostprocessor,
    'yolo_detection_head': DetectionPostprocessor,
    'rtmcc': PoseEstimationPostprocessor,
    'rtdetr_head': DetectionPostprocessor,
}



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Dict, Type

import torch
import torch.nn as nn

NORM_REGISTRY: Dict[str, Type[nn.Module]] = {
    'batch_norm': nn.BatchNorm2d,
    'instance_norm': nn.InstanceNorm2d,
}

ACTIVATION_REGISTRY: Dict[str, Type[nn.Module]] = {
    'relu': nn.ReLU,
    'relu6': nn.ReLU6,
    'prelu': nn.PReLU,
    'leaky_relu': nn.LeakyReLU,
    'gelu': nn.GELU,
    'silu': nn.SiLU,
    'swish': nn.SiLU,
    'hard_swish': nn.Hardswish,
}

POOL2D_RESGISTRY: Dict[str, Type[nn.Module]] = {
    'max': nn.MaxPool2d,
    'avg': nn.AvgPool2d,
}



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Callable, Dict, Literal, Type

from .base import BaseMetric
from .classification import ClassificationMetricAdaptor, Top1Accuracy, Top5Accuracy
from .detection import DetectionMetricAdaptor, Precision50, Recall50, mAP50, mAP50_95, mAP75
from .pose_estimation import PCK, PoseEstimationMetricAdaptor
from .segmentation import PixelAccuracy, SegmentationMetricAdaptor, mIoU

METRIC_LIST: Dict[str, Type[BaseMetric]] = {
    'top1_accuracy': Top1Accuracy,
    'top5_accuracy': Top5Accuracy,
    'miou': mIoU,
    'pixel_accuracy': PixelAccuracy,
    'precision50': Precision50,
    'recall50': Recall50,
    'map50': mAP50,
    'map75': mAP75,
    'map50_95': mAP50_95,
    'pck': PCK,
}

METRIC_ADAPTORS = {
    'classification': ClassificationMetricAdaptor,
    'segmentation': SegmentationMetricAdaptor,
    'detection': DetectionMetricAdaptor,
    'pose_estimation': PoseEstimationMetricAdaptor,
}

PHASE_LIST = ['train', 'valid', 'test']

TASK_AVAILABLE_METRICS = {
    'classification': ['top1_accuracy', 'top5_accuracy'],
    'segmentation': ['miou', 'pixel_accuracy'],
    'detection': ['precision50', 'recall50', 'map50', 'map75', 'map50_95'],
    'pose_estimation': ['pck'],
}

TASK_DEFAULT_METRICS = {
    'classification': ['top1_accuracy', 'top5_accuracy'],
    'segmentation': ['miou', 'pixel_accuracy'],
    'detection': ['map50', 'map75', 'map50_95', 'precision50', 'recall50'],
    'pose_estimation': ['pck'],
}



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import json
import os
from functools import partial
from itertools import chain
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import PIL.Image as Image
import torch.distributed as dist
from loguru import logger
from omegaconf import ListConfig

from .base import BaseCustomDataset, BaseHFDataset, BaseSampleLoader
from .utils.constants import IMG_EXTENSIONS
from .utils.misc import load_classification_class_map, natural_key

VALID_IMG_EXTENSIONS = IMG_EXTENSIONS + tuple((x.upper() for x in IMG_EXTENSIONS))


class ClassficationSampleLoader(BaseSampleLoader):
    def __init__(self, conf_data, train_valid_split_ratio):
        super(ClassficationSampleLoader, self).__init__(conf_data, train_valid_split_ratio)

    def load_data(self, split='train'):
        data_root = Path(self.conf_data.path.root)
        split_dir = self.conf_data.path[split]
        image_dir: Path = data_root / split_dir.image
        annotation_path: Optional[Path] = data_root / split_dir.label if split_dir.label is not None else None
        images_and_targets: List[Dict[str, Optional[Union[str, int]]]] = []

        assert split in ['train', 'valid', 'test'], f"split should be either {['train', 'valid', 'test']}"
        if annotation_path is not None:
            file_to_idx = load_classification_class_map(annotation_path)
            for ext in IMG_EXTENSIONS:
                for file in chain(image_dir.glob(f'*{ext}'), image_dir.glob(f'*{ext.upper()}')):
                    if file.name in file_to_idx:
                        images_and_targets.append({'image': str(file), 'label': file_to_idx[file.name]})
                        continue
                    logger.debug(f"Found file without label: {file}")
        else:
            if split in ['train', 'valid']:
                raise ValueError("For train and valid split, label path must be provided!")
            for ext in VALID_IMG_EXTENSIONS:
                images_and_targets.extend([{'image': str(file), 'label': None} for file in chain(image_dir.glob(f'*{ext}'), image_dir.glob(f'*{ext.upper()}'))])

        images_and_targets = sorted(images_and_targets, key=lambda k: natural_key(k['image']))
        return images_and_targets

    def load_id_mapping(self):
        root_path = Path(self.conf_data.path.root)

        if isinstance(self.conf_data.id_mapping, ListConfig):
            return list(self.conf_data.id_mapping)

        elif isinstance(self.conf_data.id_mapping, str):
            id_mapping_path = root_path / self.conf_data.id_mapping
            if not os.path.isfile(id_mapping_path):
                FileNotFoundError(f"File not found: {id_mapping_path}")

            with open(id_mapping_path, 'r') as f:
                id_mapping = json.load(f)
            return id_mapping

        else:
            raise ValueError(f"Unsupported id_mapping value {self.conf_data.id_mapping}")

    def load_class_map(self, id_mapping):
        idx_to_class: Dict[int, str] = dict(enumerate(id_mapping))
        return {'idx_to_class': idx_to_class}

    def load_huggingface_samples(self):
        from datasets import ClassLabel, load_dataset

        cache_dir = self.conf_data.metadata.custom_cache_dir
        root = self.conf_data.metadata.repo
        subset_name = self.conf_data.metadata.subset
        if cache_dir is not None:
            cache_dir = Path(cache_dir)
            Path(cache_dir).mkdir(exist_ok=True, parents=True)
        total_dataset = load_dataset(root, name=subset_name, cache_dir=cache_dir)

        label_feature_name = self.conf_data.metadata.features.label
        # Assumed hugging face dataset always has training split
        label_feature = total_dataset['train'].features[label_feature_name]
        if isinstance(label_feature, ClassLabel):
            labels: List[str] = label_feature.names
        else:
            labels = list({sample[label_feature_name] for sample in total_dataset['train']})

        if isinstance(labels[0], int):
            # TODO: find class_map <-> idx and apply it (ex. using id_mapping)
            idx_to_class: Dict[int, int] = {k: k for k in labels}
        elif isinstance(labels[0], str):
            idx_to_class: Dict[int, str] = dict(enumerate(labels))

        exists_valid = 'validation' in total_dataset
        exists_test = 'test' in total_dataset

        train_samples = total_dataset['train']
        valid_samples = None
        if exists_valid:
            valid_samples = total_dataset['validation']
        test_samples = None
        if exists_test:
            test_samples = total_dataset['test']

        if not exists_valid:
            splitted_datasets = train_samples.train_test_split(test_size=(1 - self.train_valid_split_ratio))
            train_samples = splitted_datasets['train']
            valid_samples = splitted_datasets['test']
        return train_samples, valid_samples, test_samples, {'idx_to_class': idx_to_class}


class ClassificationCustomDataset(BaseCustomDataset):

    def __init__(self, conf_data, conf_augmentation, model_name, idx_to_class,
                 split, samples, transform=None, **kwargs):
        super(ClassificationCustomDataset, self).__init__(
            conf_data, conf_augmentation, model_name, idx_to_class,
            split, samples, transform, **kwargs
        )

    def cache_dataset(self):
        logger.info(f'Caching | Loading samples of {self.mode} to memory... This can take minutes.')

        def _load(i, samples):
            image = Image.open(str(samples[i]['image'])).convert('RGB')
            return i, image

        num_threads = 8 # TODO: Compute appropriate num_threads
        all_indices = list(range(len(self.samples)))
        load_imgs = ThreadPool(num_threads).imap(
            partial(_load, samples=self.samples),
            all_indices
        )
        for i, image in load_imgs:
            self.samples[i]['image'] = image

        self.cache = True

    def __getitem__(self, index):
        img = self.samples[index]['image']
        target = self.samples[index]['label']
        if not self.cache:
            img = Image.open(img).convert('RGB')

        if self.transform is not None:
            out = self.transform(img)

        if target is None:
            target = -1  # To be ignored at cross-entropy loss
        return index, out['image'], target


class ClassificationHFDataset(BaseHFDataset):

    def __init__(
            self,
            conf_data,
            conf_augmentation,
            model_name,
            idx_to_class,
            split,
            huggingface_dataset,
            transform=None,
            **kwargs
    ):
        root = conf_data.metadata.repo
        super(ClassificationHFDataset, self).__init__(
            conf_data,
            conf_augmentation,
            model_name,
            root,
            split,
            transform,
        )
        # Make sure that you additionally install `requirements-data.txt`

        self.samples = huggingface_dataset
        self.idx_to_class = idx_to_class
        self.class_to_idx = {v: k for k, v in self.idx_to_class.items()}

        self.image_feature_name = conf_data.metadata.features.image
        self.label_feature_name = conf_data.metadata.features.label

    @property
    def num_classes(self):
        return len(self.idx_to_class)

    @property
    def class_map(self):
        return self.idx_to_class

    def __len__(self):
        return self.samples.num_rows

    def __getitem__(self, index):
        img: Image.Image = self.samples[index][self.image_feature_name]
        target: Union[int, str] = self.samples[index][self.label_feature_name] if self.label_feature_name in self.samples[index] else None
        if isinstance(target, str):
            target: int = self.class_to_idx[target]

        if self.transform is not None:
            out = self.transform(img)
        if target is None:
            target = -1
        return index, out['image'], target

    def cache_dataset(self):
        raise NotImplementedError("Caching is not implemented for HuggingFace datasets.")



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import warnings
from pathlib import Path
from typing import Union

import torch
from safetensors import safe_open
from safetensors.torch import save_file


def _validate_file(file_path: Path):
    file_path = Path(file_path)

    extension = file_path.suffix
    supporting_extensions = ['.pth', '.pt', '.ckpt', '.safetensors']
    assert extension in supporting_extensions, f"The checkpoint format should be one of {supporting_extensions}! The given file is {file_path}."
    return file_path, extension


def load_checkpoint(f: Union[str, Path]):
    file_path, extension = _validate_file(f)

    if extension == '.safetensors':
        state_dict = {}
        with safe_open(str(file_path), framework="pt", device='cpu') as f:
            state_dict_keys = f.keys()
            for k in state_dict_keys:
                state_dict[k] = f.get_tensor(k)
        return state_dict

    state_dict = torch.load(file_path, map_location='cpu')
    return state_dict


def save_checkpoint(obj_dict, f: Union[str, Path]) -> None:
    file_path, extension = _validate_file(f)

    if extension == '.safetensors':
        save_file(obj_dict, str(file_path))
        return

    torch.save(obj_dict, file_path)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import math
import random
from typing import List

import cv2
import numpy as np
import PIL.Image as Image
import torchvision.transforms.functional as F
from torchvision.transforms.functional import InterpolationMode


def adjust_box_anns(bbox, scale_ratio, padw, padh, w_max, h_max):
    bbox[:, 0::2] = np.clip(bbox[:, 0::2] * scale_ratio + padw, 0, w_max)
    bbox[:, 1::2] = np.clip(bbox[:, 1::2] * scale_ratio + padh, 0, h_max)
    return bbox


def get_mosaic_coordinate(mosaic_image, mosaic_index, xc, yc, w, h, input_h, input_w):
    # TODO update doc
    # index0 to top left part of image
    if mosaic_index == 0:
        x1, y1, x2, y2 = max(xc - w, 0), max(yc - h, 0), xc, yc
        small_coord = w - (x2 - x1), h - (y2 - y1), w, h
    # index1 to top right part of image
    elif mosaic_index == 1:
        x1, y1, x2, y2 = xc, max(yc - h, 0), min(xc + w, input_w * 2), yc
        small_coord = 0, h - (y2 - y1), min(w, x2 - x1), h
    # index2 to bottom left part of image
    elif mosaic_index == 2:
        x1, y1, x2, y2 = max(xc - w, 0), yc, xc, min(input_h * 2, yc + h)
        small_coord = w - (x2 - x1), 0, w, min(y2 - y1, h)
    # index2 to bottom right part of image
    elif mosaic_index == 3:
        x1, y1, x2, y2 = xc, yc, min(xc + w, input_w * 2), min(input_h * 2, yc + h)  # noqa
        small_coord = 0, 0, min(w, x2 - x1), min(y2 - y1, h)
    return (x1, y1, x2, y2), small_coord


def get_aug_params(value, center=0):
    if isinstance(value, float):
        return random.uniform(center - value, center + value)
    elif len(value) == 2:
        return random.uniform(value[0], value[1])
    else:
        raise ValueError(
            "Affine params should be either a sequence containing two values\
             or single float values. Got {}".format(value)
        )


def get_affine_matrix(
    target_size,
    degrees,
    translate,
    scales,
    shear,
):
    twidth, theight = target_size

    # Rotation and Scale
    angle = get_aug_params(degrees)
    scale = get_aug_params(scales, center=1.0)

    if scale <= 0.0:
        raise ValueError("Argument scale should be positive")

    R = cv2.getRotationMatrix2D(angle=angle, center=(0, 0), scale=scale)

    M = np.ones([2, 3])
    # Shear
    shear_x = math.tan(get_aug_params(shear) * math.pi / 180)
    shear_y = math.tan(get_aug_params(shear) * math.pi / 180)

    M[0] = R[0] + shear_y * R[1]
    M[1] = R[1] + shear_x * R[0]

    # Translation
    translation_x = get_aug_params(translate) * twidth  # x translation (pixels)
    translation_y = get_aug_params(translate) * theight  # y translation (pixels)

    M[0, 2] = translation_x
    M[1, 2] = translation_y

    return M, scale


def apply_affine_to_bboxes(targets, target_size, M, scale):
    num_gts = len(targets)

    # warp corner points
    twidth, theight = target_size
    corner_points = np.ones((4 * num_gts, 3))
    corner_points[:, :2] = targets[:, [0, 1, 2, 3, 0, 3, 2, 1]].reshape(
        4 * num_gts, 2
    )  # x1y1, x2y2, x1y2, x2y1
    corner_points = corner_points @ M.T  # apply affine transform
    corner_points = corner_points.reshape(num_gts, 8)

    # create new boxes
    corner_xs = corner_points[:, 0::2]
    corner_ys = corner_points[:, 1::2]
    new_bboxes = (
        np.concatenate(
            (corner_xs.min(1), corner_ys.min(1), corner_xs.max(1), corner_ys.max(1))
        )
        .reshape(4, num_gts)
        .T
    )

    # clip boxes
    new_bboxes[:, 0::2] = new_bboxes[:, 0::2].clip(0, twidth)
    new_bboxes[:, 1::2] = new_bboxes[:, 1::2].clip(0, theight)

    targets[:, :4] = new_bboxes

    return targets


def random_affine(
    img,
    targets,
    target_size,
    degrees,
    translate,
    scales,
    shear,
    fill,
):
    M, scale = get_affine_matrix(target_size, degrees, translate, scales, shear)

    img = cv2.warpAffine(img, M, dsize=target_size, borderValue=(fill, fill, fill))

    # Transform label coordinates
    if len(targets) > 0:
        targets = apply_affine_to_bboxes(targets, target_size, M, scale)

    return img, targets


class MosaicDetection:
    """
    Based on the MosaicDetection implementation of Megvii.
    https://github.com/Megvii-BaseDetection/YOLOX
    """
    visualize = False

    def __init__(
        self,
        size: List,
        affine_scale: List,
        mixup_scale: List,
        degrees: float,
        translate: float,
        shear: float,
        enable_mixup: bool,
        mosaic_prob: float,
        mixup_prob: float,
        fill: int,
        mosaic_off_duration: int,
    ):
        self.size = size
        self.affine_scale = affine_scale
        self.mixup_scale = mixup_scale
        self.degrees = degrees
        self.translate = translate
        self.shear = shear
        self.enable_mixup = enable_mixup
        self.mosaic_prob = mosaic_prob
        self.mixup_prob = mixup_prob
        self.fill = fill
        self.mosaic_off_duration = mosaic_off_duration

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        # Turn off mosaic augmentation when end_epoch - cur_epoch <= mosaic_off_duration
        if (self.mosaic_off_duration < (dataset.end_epoch - dataset.cur_epoch.value)) and (random.random() < self.mosaic_prob):
            mosaic_labels = []
            input_dim = self.size
            input_h, input_w = input_dim[0], input_dim[1]

            # yc, xc = s, s  # mosaic center x, y
            yc = int(random.uniform(0.5 * input_h, 1.5 * input_h))
            xc = int(random.uniform(0.5 * input_w, 1.5 * input_w))

            # 3 additional image indices
            items = [dataset.pull_item(random.randint(0, len(dataset) - 1)) for _ in range(3)]
            items = [(image, label, bbox)] + items

            c = len(image.split())
            mosaic_img = np.full((input_h * 2, input_w * 2, c), self.fill, dtype=np.uint8)

            for i_mosaic, (image, label, bbox) in enumerate(items):
                #h0, w0 = image.shape[:2]  # orig hw
                w, h = image.size
                scale = min(1. * input_h / h, 1. * input_w / w)

                image = np.array(image)
                image = cv2.resize(image, (int(w * scale), int(h * scale)), interpolation=cv2.INTER_LINEAR)

                h, w, _ = image.shape[:3]
                # suffix l means large image, while s means small image in mosaic aug.
                (l_x1, l_y1, l_x2, l_y2), (s_x1, s_y1, s_x2, s_y2) = get_mosaic_coordinate(
                    mosaic_img, i_mosaic, xc, yc, w, h, input_h, input_w
                )

                mosaic_img[l_y1:l_y2, l_x1:l_x2] = image[s_y1:s_y2, s_x1:s_x2]
                padw, padh = l_x1 - s_x1, l_y1 - s_y1

                m_labels = np.concatenate([bbox, label], axis=-1)
                # Normalized xywh to pixel xyxy format
                if m_labels.size > 0:
                    m_labels[:, 0] = scale * m_labels[:, 0] + padw
                    m_labels[:, 1] = scale * m_labels[:, 1] + padh
                    m_labels[:, 2] = scale * m_labels[:, 2] + padw
                    m_labels[:, 3] = scale * m_labels[:, 3] + padh
                mosaic_labels.append(m_labels)

            if len(mosaic_labels):
                mosaic_labels = np.concatenate(mosaic_labels, 0)
                np.clip(mosaic_labels[:, 0], 0, 2 * input_w, out=mosaic_labels[:, 0])
                np.clip(mosaic_labels[:, 1], 0, 2 * input_h, out=mosaic_labels[:, 1])
                np.clip(mosaic_labels[:, 2], 0, 2 * input_w, out=mosaic_labels[:, 2])
                np.clip(mosaic_labels[:, 3], 0, 2 * input_h, out=mosaic_labels[:, 3])

            mosaic_img, mosaic_labels = random_affine(
                mosaic_img,
                mosaic_labels,
                target_size=(input_w, input_h),
                degrees=self.degrees,
                translate=self.translate,
                scales=self.affine_scale,
                shear=self.shear,
                fill=self.fill,
            )

            # -----------------------------------------------------------------
            # CopyPaste: https://arxiv.org/abs/2012.07177
            # -----------------------------------------------------------------
            if (
                self.enable_mixup
                and len(mosaic_labels) != 0
                and random.random() < self.mixup_prob
            ):
                mosaic_img, mosaic_labels = self.mixup(mosaic_img, mosaic_labels, input_dim, dataset)

            bbox = mosaic_labels[:, :4]
            label = mosaic_labels[:, -1:]
            mosaic_img = Image.fromarray(mosaic_img) # return as PIL
            # TODO: Compute mask, keypoint
            return mosaic_img, label, mask, bbox, keypoint

        else:
            return image, label, mask, bbox, keypoint

    def mixup(self, origin_img, origin_labels, input_dim, dataset):
        jit_factor = random.uniform(*self.mixup_scale)
        FLIP = random.uniform(0, 1) > 0.5
        cp_labels = []
        while len(cp_labels) == 0:
            cp_index = random.randint(0, len(dataset) - 1)
            _, cp_labels, _ = dataset.pull_item(cp_index)
        img, cp_labels, cp_boxes = dataset.pull_item(cp_index)
        img = np.array(img)
        cp_labels = np.concatenate([cp_boxes, cp_labels], axis=-1)

        if len(img.shape) == 3:
            cp_img = np.ones((input_dim[0], input_dim[1], 3), dtype=np.uint8) * 114
        else:
            cp_img = np.ones(input_dim, dtype=np.uint8) * 114

        cp_scale_ratio = min(input_dim[0] / img.shape[0], input_dim[1] / img.shape[1])
        resized_img = cv2.resize(
            img,
            (int(img.shape[1] * cp_scale_ratio), int(img.shape[0] * cp_scale_ratio)),
            interpolation=cv2.INTER_LINEAR,
        )

        cp_img[
            : int(img.shape[0] * cp_scale_ratio), : int(img.shape[1] * cp_scale_ratio)
        ] = resized_img

        cp_img = cv2.resize(
            cp_img,
            (int(cp_img.shape[1] * jit_factor), int(cp_img.shape[0] * jit_factor)),
        )
        cp_scale_ratio *= jit_factor

        if FLIP:
            cp_img = cp_img[:, ::-1, :]

        origin_h, origin_w = cp_img.shape[:2]
        target_h, target_w = origin_img.shape[:2]
        padded_img = np.zeros(
            (max(origin_h, target_h), max(origin_w, target_w), 3), dtype=np.uint8
        )
        padded_img[:origin_h, :origin_w] = cp_img

        x_offset, y_offset = 0, 0
        if padded_img.shape[0] > target_h:
            y_offset = random.randint(0, padded_img.shape[0] - target_h - 1)
        if padded_img.shape[1] > target_w:
            x_offset = random.randint(0, padded_img.shape[1] - target_w - 1)
        padded_cropped_img = padded_img[
            y_offset: y_offset + target_h, x_offset: x_offset + target_w
        ]

        cp_bboxes_origin_np = adjust_box_anns(
            cp_labels[:, :4].copy(), cp_scale_ratio, 0, 0, origin_w, origin_h
        )
        if FLIP:
            cp_bboxes_origin_np[:, 0::2] = (
                origin_w - cp_bboxes_origin_np[:, 0::2][:, ::-1]
            )
        cp_bboxes_transformed_np = cp_bboxes_origin_np.copy()
        cp_bboxes_transformed_np[:, 0::2] = np.clip(
            cp_bboxes_transformed_np[:, 0::2] - x_offset, 0, target_w
        )
        cp_bboxes_transformed_np[:, 1::2] = np.clip(
            cp_bboxes_transformed_np[:, 1::2] - y_offset, 0, target_h
        )

        cls_labels = cp_labels[:, 4:5].copy()
        box_labels = cp_bboxes_transformed_np
        labels = np.hstack((box_labels, cls_labels))
        origin_labels = np.vstack((origin_labels, labels))
        origin_img = origin_img.astype(np.float32)
        origin_img = 0.5 * origin_img + 0.5 * padded_cropped_img.astype(np.float32)

        return origin_img.astype(np.uint8), origin_labels

    def __repr__(self) -> str:
        return "{}(size={}, mosaic_prob={}, affine_scale={}, degrees={}, translate={}, shear={}, enable_mixup={}, mixup_prob={}, mixup_scale={}, fill={})".format(
            self.__class__.__name__, self.size, self.mosaic_prob, self.affine_scale, self.degrees, self.translate, self.shear, self.enable_mixup, self.mixup_prob, self.mixup_scale, self.fill
        )



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Dict, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor


class CrossEntropyLoss(nn.Module):
    def __init__(self, weight: Optional[Tensor]=None, size_average=None, ignore_index: int=-100,
                 reduce=None, label_smoothing: float=0.0, **kwargs):
        super(CrossEntropyLoss, self).__init__()
        self.loss_fn = nn.CrossEntropyLoss(weight=weight, size_average=size_average, ignore_index=ignore_index,
                                           reduce=reduce, reduction='mean', label_smoothing=label_smoothing)

    def forward(self, out: Dict, target: Dict) -> Dict:
        pred = out['pred']
        target = target['target']
        loss = self.loss_fn(pred, target)
        return loss


class SigmoidFocalLoss(nn.Module):
    def __init__(self, alpha, gamma, **kwargs):
        super(SigmoidFocalLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma

    def forward(self, out: Dict, target: Dict, reduction='mean'):
        pred = out['pred']
        target = target['target']
        assert pred.shape == target.shape, 'Tensor shapes of prediction and target must be same for SigmoidFocalLoss.'

        p = torch.sigmoid(pred)
        ce_loss = F.binary_cross_entropy_with_logits(pred, target, reduction="none")
        p_t = p * target + (1 - p) * (1 - target)
        loss = ce_loss * ((1 - p_t) ** self.gamma)

        if self.alpha >= 0:
            alpha_t = self.alpha * target + (1 - self.alpha) * (1 - target)
            loss = alpha_t * loss

        if reduction == 'none':
            pass
        elif reduction == 'mean':
            loss = loss.mean()
        elif reduction == 'sum':
            loss = loss.sum()
        else:
            raise ValueError(
                f"Invalid Value for arg 'reduction': '{reduction} \n Supported reduction modes: 'none', 'mean', 'sum'"
            )
        return loss



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import math
import random
from collections.abc import Sequence
from typing import Any, Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
import PIL.Image as Image
import torch
import torchvision.transforms as T
import torchvision.transforms.functional as F
from torch import Tensor
from torchvision.ops.boxes import box_iou
from torchvision.transforms.autoaugment import _apply_op
from torchvision.transforms.functional import InterpolationMode
from torchvision.transforms.transforms import _check_sequence_input

BBOX_CROP_KEEP_THRESHOLD = 0.2
MAX_RETRY = 5
INVERSE_MODES_MAPPING = {
    'nearest': InterpolationMode.NEAREST,
    'bilinear': InterpolationMode.BILINEAR,
    'bicubic': InterpolationMode.BICUBIC,
}


class Compose:
    def __init__(self, transforms, additional_targets: Dict = None):
        if additional_targets is None:
            additional_targets = {}
        self.transforms = transforms
        self.additional_targets = additional_targets

    def _get_transformed(self, image, label, mask, bbox, keypoint, visualize_for_debug, dataset):
        for t in self.transforms:
            if visualize_for_debug and not t.visualize:
                continue
            image, label, mask, bbox, keypoint = t(image=image, label=label, mask=mask, bbox=bbox, keypoint=keypoint, dataset=dataset)
        return image, label, mask, bbox, keypoint

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, visualize_for_debug=False, dataset=None, **kwargs):
        additional_targets_result = {k: None for k in kwargs if k in self.additional_targets}

        result_image, result_label, result_mask, result_bbox, result_keypoint = self._get_transformed(image=image, label=label, mask=mask, bbox=bbox, keypoint=keypoint, dataset=dataset, visualize_for_debug=visualize_for_debug)
        for key in additional_targets_result:
            if self.additional_targets[key] == 'mask':
                _, _, additional_targets_result[key], _, _ = self._get_transformed(image=image, label=label, mask=kwargs[key], bbox=None, keypoint=keypoint, dataset=dataset, visualize_for_debug=visualize_for_debug)
            elif self.additional_targets[key] == 'bbox':
                _, _, _, additional_targets_result[key], _ = self._get_transformed(image=image, label=label, mask=None, bbox=kwargs[key], keypoint=keypoint, dataset=dataset, visualize_for_debug=visualize_for_debug)
            else:
                del additional_targets_result[key]

        return_dict = {'image': result_image, 'label': result_label}
        if mask is not None:
            return_dict.update({'mask': result_mask})
        if bbox is not None:
            return_dict.update({'bbox': result_bbox})
        if keypoint is not None:
            return_dict.update({'keypoint': result_keypoint})
        return_dict.update(additional_targets_result)
        return return_dict

    def __repr__(self):
        compose_summary = "CustomCompose"
        compose_list = ",\n\t".join([str(t) for t in self.transforms])
        compose_summary += "(\n\t" + compose_list + "\n)"
        return compose_summary


class CenterCrop(T.CenterCrop):
    visualize = True

    def __init__(
        self,
        size: Union[int, List],
    ):
        super().__init__(size)

    def forward(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        # TODO: Compute mask, bbox, keypoint
        return F.center_crop(image, self.size), label, mask, bbox, keypoint

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(size={self.size})"


class Identity:
    visualize = True

    def __init__(self):
        pass

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        return image, label, mask, bbox, keypoint

    def __repr__(self):
        return self.__class__.__name__ + "()"


class Resize(T.Resize):
    visualize = True

    def __init__(
        self,
        size: Union[int, List],
        interpolation: str,
        max_size: Optional[int],
        resize_criteria: Optional[str],
    ):
        if isinstance(size, int):
            assert resize_criteria is not None, "If size is int, resize_criteria must be 'short' or 'long'"
        self.resize_criteria = resize_criteria
        interpolation = INVERSE_MODES_MAPPING[interpolation]
        # @illian01: antialias paramter always true (always use) for PIL image, and NetsPresso Trainer uses PIL image for augmentation.
        super().__init__(size, interpolation, max_size)

    def forward(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        w, h = image.size

        if isinstance(self.size, int) and self.resize_criteria == 'long':
            long_side, short_side = max(h, w), min(h, w)
            resize_factor = self.size / long_side
            target_size = [self.size, round(resize_factor * short_side)] if h > w else [round(resize_factor * short_side), self.size]
        else:
            target_size = self.size

        image = F.resize(image, target_size, self.interpolation, self.max_size, self.antialias)
        if mask is not None:
            mask = F.resize(mask, target_size, interpolation=T.InterpolationMode.NEAREST,
                            max_size=self.max_size)
        if bbox is not None:
            target_w, target_h = image.size # @illian01: Determine ratio according to the actual resized image
            bbox[..., 0:4:2] *= float(target_w / w)
            bbox[..., 1:4:2] *= float(target_h / h)
        # TODO: Compute keypoint
        return image, label, mask, bbox, keypoint

    def __repr__(self):
        return self.__class__.__name__ + "(size={0}, interpolation={1}, max_size={2}, antialias={3}, resize_criteria={4})".format(
            self.size, self.interpolation.value, self.max_size, self.antialias, self.resize_criteria)


class RandomHorizontalFlip:
    visualize = True

    def __init__(
        self,
        p: float,
    ):
        self.p: float = max(0., min(1., p))

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        w, _ = image.size
        if random.random() < self.p:
            image = F.hflip(image)
            if mask is not None:
                mask = F.hflip(mask)
            if bbox is not None:
                bbox[..., 2::-2] = w - bbox[..., 0:4:2]
            if keypoint is not None:
                keypoint = keypoint[:, dataset.flip_indices] # flip_indices must be defined in config (swap)
                keypoint[..., 0] = w - keypoint[..., 0]
        return image, label, mask, bbox, keypoint

    def __repr__(self):
        return self.__class__.__name__ + "(p={0})".format(self.p)


class RandomVerticalFlip:
    visualize = True

    def __init__(
        self,
        p: float,
    ):
        self.p: float = max(0., min(1., p))

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        _, h = image.size
        if random.random() < self.p:
            image = F.vflip(image)
            if mask is not None:
                mask = F.vflip(mask)
            if bbox is not None:
                bbox[..., 3::-2] = h - bbox[..., 1:4:2]
            if keypoint is not None:
                keypoint = keypoint[:, dataset.flip_indices] # flip_indices must be defined in config (swap)
                keypoint[..., 1] = h - keypoint[..., 1]
        return image, label, mask, bbox, keypoint

    def __repr__(self):
        return self.__class__.__name__ + "(p={0})".format(self.p)


class Pad:
    visualize = True

    def __init__(
        self,
        size: Union[int, List],
        fill: Union[int, List],
    ):
        super().__init__()
        if not isinstance(size, (int, Sequence)):
            raise TypeError("Size should be int or sequence. Got {}".format(type(size)))
        if isinstance(size, Sequence) and len(size) != 2:
            raise ValueError("If size is a sequence, it should have 2 values")
        self.new_h = size[0] if isinstance(size, Sequence) else size
        self.new_w = size[1] if isinstance(size, Sequence) else size
        self.fill = fill
        self.padding_mode = 'constant' # @illian: Fix as constant. I think other options are not gonna used well.

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        if not isinstance(image, (torch.Tensor, Image.Image)):
            raise TypeError("Image should be Tensor or PIL.Image. Got {}".format(type(image)))

        if isinstance(image, Image.Image):
            w, h = image.size
        else:
            w, h = image.shape[-1], image.shape[-2]

        w_pad_needed = max(0, self.new_w - w)
        h_pad_needed = max(0, self.new_h - h)
        # @illian01: I think we don't need to pad in all direction.
        padding_ltrb = [0, 0, w_pad_needed, h_pad_needed]
        '''
        padding_ltrb = [w_pad_needed // 2,
                        h_pad_needed // 2,
                        w_pad_needed // 2 + w_pad_needed % 2,
                        h_pad_needed // 2 + h_pad_needed % 2]
        '''
        image = F.pad(image, padding_ltrb, fill=self.fill, padding_mode=self.padding_mode)
        if mask is not None:
            mask = F.pad(mask, padding_ltrb, fill=255, padding_mode=self.padding_mode)
        if bbox is not None:
            padding_left, padding_top, _, _ = padding_ltrb
            bbox[..., 0:4:2] += padding_left
            bbox[..., 1:4:2] += padding_top
        # TODO: Compute keypoint
        return image, label, mask, bbox, keypoint

    def __repr__(self):
        return self.__class__.__name__ + "(size={0}, fill={1}, padding_mode={2})".format(
            (self.new_h, self.new_w), self.fill, self.padding_mode
        )


class ColorJitter(T.ColorJitter):
    visualize = True

    def __init__(
        self,
        brightness: Union[float, List],
        contrast: Union[float, List],
        saturation: Union[float, List],
        hue: Union[float, List],
        p: float
    ):
        super(ColorJitter, self).__init__(brightness, contrast, saturation, hue)
        self.p: float = max(0., min(1., p))

    def forward(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        fn_idx, brightness_factor, contrast_factor, saturation_factor, hue_factor = \
            self.get_params(self.brightness, self.contrast, self.saturation, self.hue)

        if random.random() < self.p:
            for fn_id in fn_idx:
                if fn_id == 0 and brightness_factor is not None:
                    image = F.adjust_brightness(image, brightness_factor)
                elif fn_id == 1 and contrast_factor is not None:
                    image = F.adjust_contrast(image, contrast_factor)
                elif fn_id == 2 and saturation_factor is not None:
                    image = F.adjust_saturation(image, saturation_factor)
                elif fn_id == 3 and hue_factor is not None:
                    image = F.adjust_hue(image, hue_factor)

        return image, label, mask, bbox, keypoint

    def __repr__(self):
        return self.__class__.__name__ + \
            f"(brightness={self.brightness}, " + \
            f"contrast={self.contrast}, " + \
            f"saturation={self.saturation}, " + \
            f"hue={self.hue}, " + \
            f"p={self.p})"


class RandomCrop:
    visualize = True

    def __init__(
        self,
        size: Union[int, List],
        fill: int,
    ):

        if not isinstance(size, (int, Sequence)):
            raise TypeError("Size should be int or sequence. Got {}".format(type(size)))
        if isinstance(size, Sequence) and len(size) not in (1, 2):
            raise ValueError("If size is a sequence, it should have 1 or 2 values")
        self.size_h = size[0] if isinstance(size, Sequence) else size
        self.size_w = size[1] if isinstance(size, Sequence) else size
        self.image_pad_if_needed = Pad((self.size_h, self.size_w), fill=fill)

    def _crop_bbox(self, bbox, i, j, h, w):
        area_original = (bbox[..., 2] - bbox[..., 0]) * (bbox[..., 3] - bbox[..., 1])

        bbox[..., 0:4:2] = np.clip(bbox[..., 0:4:2] - j, 0, w)
        bbox[..., 1:4:2] = np.clip(bbox[..., 1:4:2] - i, 0, h)

        area_cropped = (bbox[..., 2] - bbox[..., 0]) * (bbox[..., 3] - bbox[..., 1])
        area_ratio = area_cropped / (area_original + 1)  # +1 for preventing ZeroDivisionError

        bbox = bbox[area_ratio >= BBOX_CROP_KEEP_THRESHOLD, ...]
        return bbox

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        image, _, mask, bbox, _ = self.image_pad_if_needed(image=image, mask=mask, bbox=bbox)
        i, j, h, w = T.RandomCrop.get_params(image, (self.size_h, self.size_w))
        image = F.crop(image, i, j, h, w)
        if mask is not None:
            mask = F.crop(mask, i, j, h, w)
        if bbox is not None:
            bbox_candidate = self._crop_bbox(bbox, i, j, h, w)
            _bbox_crop_count = 1
            while bbox_candidate.shape[0] == 0:
                if _bbox_crop_count == MAX_RETRY:
                    raise ValueError(f"It seems no way to use crop augmentation for this dataset. bbox: {bbox}, (i, j, h, w): {(i, j, h, w)}")
                bbox_candidate = self._crop_bbox(bbox, i, j, h, w)
                _bbox_crop_count += 1
            bbox = bbox_candidate
        # TODO: Compute keypoint
        return image, label, mask, bbox, keypoint

    def __repr__(self):
        return self.__class__.__name__ + "(size={0})".format((self.size_h, self.size_w))


class RandomResizedCrop(T.RandomResizedCrop):
    visualize = True

    def __init__(
        self,
        size: Union[int, List],
        scale: Union[float, List],
        ratio: Union[float, List],
        interpolation: str,
    ):
        interpolation = INVERSE_MODES_MAPPING[interpolation]
        # @illian01: antialias paramter always true (always use) for PIL image, and NetsPresso Trainer uses PIL image for augmentation.
        super().__init__(size, scale, ratio, interpolation)

    def _crop_bbox(self, bbox, i, j, h, w):
        area_original = (bbox[..., 2] - bbox[..., 0]) * (bbox[..., 3] - bbox[..., 1])

        bbox[..., 0:4:2] = np.clip(bbox[..., 0:4:2] - j, 0, w)
        bbox[..., 1:4:2] = np.clip(bbox[..., 1:4:2] - i, 0, h)

        area_cropped = (bbox[..., 2] - bbox[..., 0]) * (bbox[..., 3] - bbox[..., 1])
        area_ratio = area_cropped / (area_original + 1)  # +1 for preventing ZeroDivisionError

        bbox = bbox[area_ratio >= BBOX_CROP_KEEP_THRESHOLD, ...]
        return bbox

    def forward(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        w_orig, h_orig = image.size
        i, j, h, w = self.get_params(image, self.scale, self.ratio)
        image = F.resized_crop(image, i, j, h, w, self.size, self.interpolation)
        if mask is not None:
            mask = F.resized_crop(mask, i, j, h, w, self.size, interpolation=T.InterpolationMode.NEAREST)
        if bbox is not None:
            # img = crop(img, top, left, height, width)
            bbox_candidate = self._crop_bbox(bbox, i, j, h, w)
            _bbox_crop_count = 1
            while bbox_candidate.shape[0] != 0:
                if _bbox_crop_count == MAX_RETRY:
                    raise ValueError(f"It seems no way to use crop augmentation for this dataset. bbox: {bbox}, (i, j, h, w): {(i, j, h, w)}")
                bbox_candidate = self._crop_bbox(bbox, i, j, h, w)
                _bbox_crop_count += 1
            bbox = bbox_candidate

            # img = resize(img, size, interpolation)
            w_cropped, h_cropped = np.clip(w_orig - j, 0, w), np.clip(h_orig - i, 0, h)
            target_w, target_h = (self.size, self.size) if isinstance(self.size, int) else self.size
            bbox[..., 0:4:2] *= float(target_w / w_cropped)
            bbox[..., 1:4:2] *= float(target_h / h_cropped)

        return image, label, mask, bbox, keypoint

    def __repr__(self):
        interpolate_str = self.interpolation.value
        format_string = self.__class__.__name__ + '(size={0}'.format(self.size)
        format_string += ', scale={0}'.format(tuple(round(s, 4) for s in self.scale))
        format_string += ', ratio={0}'.format(tuple(round(r, 4) for r in self.ratio))
        format_string += ', interpolation={0})'.format(interpolate_str)
        return format_string


class RandomErasing(T.RandomErasing):
    visualize = True

    def __init__(
        self,
        p: float,
        scale: List,
        ratio: List,
        value: Optional[int],
        inplace: bool,
    ):
        scale = list(scale)
        ratio = list(ratio)
        super().__init__(p, scale, ratio, value, inplace)

    @staticmethod
    def get_params(
        img, scale: Tuple[float, float], ratio: Tuple[float, float], value: Optional[int] = None
    ):
        img_w, img_h = img.size

        area = img_h * img_w

        log_ratio = torch.log(torch.tensor(ratio))
        for _ in range(10):
            erase_area = area * torch.empty(1).uniform_(scale[0], scale[1]).item()
            aspect_ratio = torch.exp(torch.empty(1).uniform_(log_ratio[0], log_ratio[1])).item()

            h = int(round(math.sqrt(erase_area * aspect_ratio)))
            w = int(round(math.sqrt(erase_area / aspect_ratio)))
            if not (h < img_h and w < img_w):
                continue

            if value is None:
                v = np.random.randint(255, size=(h, w)).astype('uint8')
                v = Image.fromarray(v).convert(img.mode)
            else:
                v = Image.new(img.mode, (w, h), value)

            i = torch.randint(0, img_h - h + 1, size=(1,)).item()
            j = torch.randint(0, img_w - w + 1, size=(1,)).item()
            return i, j, v

        # Return original image
        return 0, 0, img

    def forward(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        if torch.rand(1) < self.p:
            x, y, v = self.get_params(image, scale=self.scale, ratio=self.ratio, value=self.value)
            image.paste(v, (y, x))
            # TODO: Object-aware
            return image, label, mask, bbox, keypoint
        return image, label, mask, bbox, keypoint


class RandomIoUCrop:
    """
    Based on the torchvision implementation.
    https://pytorch.org/vision/stable/_modules/torchvision/transforms/v2/_geometry.html#RandomIoUCrop
    """
    visualize = True
    def __init__(
        self,
        min_scale: float,
        max_scale: float,
        min_aspect_ratio: float,
        max_aspect_ratio: float,
        p: float,
        sampler_options: Optional[List[float]] = None,
        trials: int = 40,
    ):
        self.min_scale = min_scale
        self.max_scale = max_scale
        self.min_aspect_ratio = min_aspect_ratio
        self.max_aspect_ratio = max_aspect_ratio
        if sampler_options is None:
            sampler_options = [0.0, 0.1, 0.3, 0.5, 0.7, 0.9, 1.0]
        self.options = sampler_options
        self.trials = trials
        self.p = p

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        if not isinstance(image, (torch.Tensor, Image.Image)):
            raise TypeError("Image should be Tensor or PIL.Image. Got {}".format(type(image)))

        if isinstance(image, Image.Image):
            w, h = image.size
        else:
            w, h = image.shape[-1], image.shape[-2]
        if random.random() < self.p:
            while True:
                idx = int(torch.randint(low=0, high=len(self.options), size=(1,)))
                min_jaccard_overlap = self.options[idx]
                if min_jaccard_overlap >= 1.0:
                    return image, label, mask, bbox, keypoint
                for _ in range(self.trials):
                    r = self.min_scale + (self.max_scale - self.min_scale) * torch.rand(2)
                    new_w = int(w * r[0])
                    new_h = int(h * r[1])
                    aspect_ratio = new_w / new_h
                    if not (self.min_aspect_ratio <= aspect_ratio <= self.max_aspect_ratio):
                        continue

                    # check for 0 area crops
                    r = torch.rand(2)
                    left = int((w - new_w) * r[0])
                    top = int((h - new_h) * r[1])
                    right = left + new_w
                    bottom = top + new_h
                    if left == right or top == bottom:
                        continue
                    xyxy_bboxes = bbox
                    cx = 0.5 * (xyxy_bboxes[..., 0] + xyxy_bboxes[..., 2])
                    cy = 0.5 * (xyxy_bboxes[..., 1] + xyxy_bboxes[..., 3])
                    is_within_crop_area = (left < cx) & (cx < right) & (top < cy) & (cy < bottom)
                    if not is_within_crop_area.any():
                        continue
                    xyxy_bboxes = torch.tensor(xyxy_bboxes[is_within_crop_area])
                    ious = box_iou(
                        xyxy_bboxes,
                        torch.tensor([[left, top, right, bottom]], dtype=xyxy_bboxes.dtype,
                                     device=xyxy_bboxes.device),)
                    if ious.max() < min_jaccard_overlap:
                        continue
                    image = F.crop(image, top, left, new_h, new_w)
                    if bbox is not None:
                        bbox = self._crop_bbox(bbox, top, left, new_h, new_w)
                    return image, label, mask, bbox, keypoint

        return image, label, mask, bbox, keypoint

    def _crop_bbox(self, bbox, i, j, h, w):
        bbox[..., 0:4:2] = np.clip(bbox[..., 0:4:2] - j, 0, w)
        bbox[..., 1:4:2] = np.clip(bbox[..., 1:4:2] - i, 0, h)
        return bbox

class RandomZoomOut:
    """
    Based on the torchvision implementation.
    https://pytorch.org/vision/0.18/_modules/torchvision/transforms/v2/_geometry.html#RandomZoomOut
    """
    visualize = True

    def __init__(
        self,
        fill: int,
        side_range: List[float],
        p: float,

    ) -> None:
        _check_sequence_input(side_range, "side_range", req_sizes=(2,))
        if side_range[0] < 1.0 or side_range[0] > side_range[1]:
            raise ValueError("Invalid side range provided {}.".format(side_range))
        self.fill = fill
        self.padding_mode = 'constant'
        self.side_range = side_range
        self.p = p

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        if not isinstance(image, (torch.Tensor, Image.Image)):
            raise TypeError("Image should be Tensor or PIL.Image. Got {}".format(type(image)))

        if isinstance(image, Image.Image):
            w, h = image.size
        else:
            w, h = image.shape[-1], image.shape[-2]

        if random.random() < self.p:
            r = self.side_range[0] + torch.rand(1) * (self.side_range[1] - self.side_range[0])
            canvas_width = int(w * r)
            canvas_height = int(h * r)

            r = torch.rand(2)
            left= int((canvas_width - w) * r[0])
            top = int((canvas_height - h) * r[1])
            right = canvas_width - (left + w)
            bottom = canvas_height - (top + h)
            padding_ltrb = [left, top, right, bottom]
            image = F.pad(image, padding_ltrb, fill=self.fill, padding_mode=self.padding_mode)
            if mask is not None:
                mask = F.pad(mask, padding_ltrb, fill=255, padding_mode=self.padding_mode)
            if bbox is not None:
                padding_left, padding_top, _, _ = padding_ltrb
                bbox[..., 0:4:2] += padding_left
                bbox[..., 1:4:2] += padding_top
            #TODO: Compute Keypoint

        return image, label, mask, bbox, keypoint


class TrivialAugmentWide(torch.nn.Module):
    """
    Based on the torchvision implementation.
    https://pytorch.org/vision/main/_modules/torchvision/transforms/autoaugment.html#TrivialAugmentWide
    """
    visualize = True

    def __init__(
        self,
        num_magnitude_bins: int,
        interpolation: str,
        fill: Optional[Union[List[int], int]],
    ) -> None:
        super().__init__()
        interpolation = INVERSE_MODES_MAPPING[interpolation]

        self.num_magnitude_bins = num_magnitude_bins
        self.interpolation = interpolation
        self.fill = fill

    def _augmentation_space(self, num_bins: int) -> Dict[str, Tuple[Tensor, bool]]:
        return {
            # op_name: (magnitudes, signed)
            "Identity": (torch.tensor(0.0), False),
            "ShearX": (torch.linspace(0.0, 0.99, num_bins), True),
            "ShearY": (torch.linspace(0.0, 0.99, num_bins), True),
            "TranslateX": (torch.linspace(0.0, 32.0, num_bins), True),
            "TranslateY": (torch.linspace(0.0, 32.0, num_bins), True),
            "Rotate": (torch.linspace(0.0, 135.0, num_bins), True),
            "Brightness": (torch.linspace(0.0, 0.99, num_bins), True),
            "Color": (torch.linspace(0.0, 0.99, num_bins), True),
            "Contrast": (torch.linspace(0.0, 0.99, num_bins), True),
            "Sharpness": (torch.linspace(0.0, 0.99, num_bins), True),
            "Posterize": (8 - (torch.arange(num_bins) / ((num_bins - 1) / 6)).round().int(), False),
            "Solarize": (torch.linspace(255.0, 0.0, num_bins), False),
            "AutoContrast": (torch.tensor(0.0), False),
            "Equalize": (torch.tensor(0.0), False),
        }

    def forward(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        fill = self.fill
        channels, height, width = F.get_dimensions(image)
        if isinstance(image, Tensor):
            if isinstance(fill, (int, float)):
                fill = [float(fill)] * channels
            elif fill is not None:
                fill = [float(f) for f in fill]

        op_meta = self._augmentation_space(self.num_magnitude_bins)
        op_index = int(torch.randint(len(op_meta), (1,)).item())
        op_name = list(op_meta.keys())[op_index]
        magnitudes, signed = op_meta[op_name]
        magnitude = (
            float(magnitudes[torch.randint(len(magnitudes), (1,), dtype=torch.long)].item())
            if magnitudes.ndim > 0
            else 0.0
        )
        if signed and torch.randint(2, (1,)):
            magnitude *= -1.0

        # TODO: Compute mask, bbox
        return _apply_op(image, op_name, magnitude, interpolation=self.interpolation, fill=fill), label, mask, bbox, keypoint

    def __repr__(self) -> str:
        s = (
            f"{self.__class__.__name__}("
            f"num_magnitude_bins={self.num_magnitude_bins}"
            f", interpolation={self.interpolation}"
            f", fill={self.fill}"
            f")"
        )
        return s


class AutoAugment(T.AutoAugment):
    visualize = True

    def __init__(
        self,
        policy: str,
        interpolation: str,
        fill: Optional[List[float]],
    ) -> None:
        policy = T.AutoAugmentPolicy(policy)
        interpolation = INVERSE_MODES_MAPPING[interpolation]

        super().__init__(policy, interpolation, fill)

    def forward(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        """
            img (PIL Image or Tensor): Image to be transformed.

        Returns:
            PIL Image or Tensor: AutoAugmented image.
        """
        fill = self.fill
        channels, height, width = F.get_dimensions(image)
        if isinstance(image, Tensor):
            if isinstance(fill, (int, float)):
                fill = [float(fill)] * channels
            elif fill is not None:
                fill = [float(f) for f in fill]

        transform_id, probs, signs = self.get_params(len(self.policies))

        op_meta = self._augmentation_space(10, (height, width))
        for i, (op_name, p, magnitude_id) in enumerate(self.policies[transform_id]):
            if probs[i] <= p:
                magnitudes, signed = op_meta[op_name]
                magnitude = float(magnitudes[magnitude_id].item()) if magnitude_id is not None else 0.0
                if signed and signs[i] == 0:
                    magnitude *= -1.0
                image = _apply_op(image, op_name, magnitude, interpolation=self.interpolation, fill=fill)

        # TODO: Compute mask, bbox
        return image, label, mask, bbox, keypoint


class HSVJitter:
    visualize = True

    def __init__(
        self,
        h_mag: int,
        s_mag: int,
        v_mag: int,
    ):
        self.h_mag = h_mag
        self.s_mag = s_mag
        self.v_mag = v_mag

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        hsv_augs = np.random.uniform(-1, 1, 3) * [self.h_mag, self.s_mag, self.v_mag]  # random gains
        hsv_augs *= np.random.randint(0, 2, 3)  # random selection of h, s, v
        hsv_augs = hsv_augs.astype(np.int16)

        img_hsv = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2HSV).astype(np.int16)

        img_hsv[..., 0] = (img_hsv[..., 0] + hsv_augs[0]) % 180
        img_hsv[..., 1] = np.clip(img_hsv[..., 1] + hsv_augs[1], 0, 255)
        img_hsv[..., 2] = np.clip(img_hsv[..., 2] + hsv_augs[2], 0, 255)

        image = cv2.cvtColor(img_hsv.astype(np.uint8), cv2.COLOR_HSV2RGB)
        image = Image.fromarray(image)
        return image, label, mask, bbox, keypoint

    def __repr__(self):
        return self.__class__.__name__ + "(h_mag={0}, s_mag={1}, v_mag={2})".format(
            self.h_mag, self.s_mag, self.v_mag)


class RandomResize:
    visualize = True

    def __init__(
        self,
        base_size: List,
        stride: int,
        random_range: int,
        interpolation: str,
    ):
        self.base_size = base_size
        self.stride = stride
        self.random_range = random_range
        self.resize = Resize(size=base_size, interpolation=interpolation, max_size=None, resize_criteria=None)
        self.counter = 0

    def random_set(self):
        delta = self.stride * random.randint(-self.random_range, self.random_range)
        size = [self.base_size[0] + delta, self.base_size[1] + delta]
        self.resize.size = size

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        """
        @illian01:
            Count random resized samples.
            If one batch completed, randomly reset target size and set counter to 0.
        """
        if self.counter == dataset.batch_size:
            self.random_set()
            self.counter = 0
        image, label, mask, bbox, keypoint = self.resize(image, label, mask, bbox, dataset)
        self.counter += 1
        return image, label, mask, bbox, keypoint

    def __repr__(self):
        return self.__class__.__name__ + "(base_size={0}, stride={1}, random_range={2})".format(
            self.base_size, self.stride, self.random_range
        )


class RandomResize2:
    visualize = True

    def __init__(
        self,
        base_size: List[int],
        random_range: Union[int, float],
        interpolation: str,
    ):
        self.base_size = base_size
        self.random_range = random_range
        # Temporarily assign 0 size to Resize class
        self.resize = Resize(size=self.base_size, interpolation=interpolation, max_size=None, resize_criteria=None)

    def random_set(self, image):
        target_h, target_w = self.base_size
        resize_factor = random.random() * (self.random_range[1] - self.random_range[0]) + self.random_range[0]
        max_size = (int(target_w * resize_factor), int(target_h * resize_factor))
        w, h = image.size
        resize_factor = min(max_size[0] / w, max_size[1] / h)
        size = [int(h * resize_factor), int(w * resize_factor)]
        self.resize.size = size

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        self.random_set(image)
        image, label, mask, bbox, keypoint = self.resize(image, label, mask, bbox, keypoint, dataset)
        return image, label, mask, bbox, keypoint

    def __repr__(self):
        return self.__class__.__name__ + "(max_size={0}, random_range={1})".format(
            self.max_size, self.random_range
        )


class PoseTopDownAffine:
    """
    Based on the mmpose implementation.
    https://github.com/open-mmlab/mmpose
    """
    visualize = False

    def __init__(
        self,
        scale: List,
        scale_prob: float,
        translate: float,
        translate_prob: float,
        rotation: int,
        rotation_prob: float,
        size: List,
    ):
        self.scale = scale
        self.scale_prob = scale_prob
        self.translate = translate
        self.translate_prob = translate_prob
        self.rotation = rotation
        self.rotation_prob = rotation_prob
        self.size = size

        assert len(self.size) == 2, "Target size must be a list of length 2"

    def trunc_normal_(self, low: float, high: float, size: Tuple = (1)):
        """
        Copied from torch.nn.init._no_grad_trunc_normal_
        This is instead of scipy.stats.tuncnorm (Not to add dependency)
        """
        mean = 0.
        std = 1.
        # Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
        def norm_cdf(x):
            # Computes standard normal cumulative distribution function
            return (1. + math.erf(x / math.sqrt(2.))) / 2.

        with torch.no_grad():
            tensor = torch.zeros(size).to(torch.float32)

            lower = norm_cdf((low - mean) / std)
            upper = norm_cdf((high - mean) / std)

            tensor.uniform_(2 * lower - 1, 2 * upper - 1)

            tensor.erfinv_()

            tensor.mul_(std * math.sqrt(2.))
            tensor.add_(mean)

            tensor.clamp_(min=low, max=high)
            return tensor.numpy()

    def _rotate_point(self, pt: np.ndarray, angle_rad: float):
        """
        Rotate a point by an angle.
        """
        sn, cs = np.sin(angle_rad), np.cos(angle_rad)
        rot_mat = np.array([[cs, -sn], [sn, cs]])
        return rot_mat @ pt

    def _get_3rd_point(self, a: np.ndarray, b: np.ndarray):
        """
        To calculate the affine matrix, three pairs of points are required. This
        function is used to get the 3rd point, given 2D points a & b.
        """
        direction = a - b
        c = b + np.r_[-direction[1], direction[0]]
        return c

    def get_warp_matrix(self, box_center: np.ndarray, box_wh: np.ndarray, rot: float):
        """
        Calculate the affine transformation matrix that can warp the bbox area
        in the input image to the output size.
        """
        shift = (0, 0) # Fix as 0
        assert len(box_center) == 2
        assert len(box_wh) == 2
        assert len(shift) == 2

        shift = np.array(shift)
        src_w, src_h = box_wh[:2]
        dst_w, dst_h = self.size

        rot_rad = np.deg2rad(rot)
        src_dir = self._rotate_point(np.array([src_w * -0.5, 0.]), rot_rad)
        dst_dir = np.array([dst_w * -0.5, 0.])

        src = np.zeros((3, 2), dtype=np.float32)
        src[0, :] = box_center + box_wh * shift
        src[1, :] = box_center + src_dir + box_wh * shift

        dst = np.zeros((3, 2), dtype=np.float32)
        dst[0, :] = [dst_w * 0.5, dst_h * 0.5]
        dst[1, :] = np.array([dst_w * 0.5, dst_h * 0.5]) + dst_dir

        src[2, :] = self._get_3rd_point(src[0, :], src[1, :])
        dst[2, :] = self._get_3rd_point(dst[0, :], dst[1, :])

        warp_mat = cv2.getAffineTransform(np.float32(src), np.float32(dst))
        return warp_mat

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        image = np.array(image)

        # This is only for one instance (bbox)
        bbox_ = bbox[0].reshape(2, 2).copy()
        bbox_center = bbox_.sum(axis=0) / 2.
        bbox_wh = bbox_[1] - bbox_[0]
        bbox_ = bbox_.reshape(-1)

        # Randomly scale, shift box, and get random rotation degree which is applied in affine transform
        scale_min, scale_max = self.scale
        mu = (scale_max + scale_min) * 0.5
        sigma = (scale_max - scale_min) * 0.5
        scale = self.trunc_normal_(low=-1., high=1., size=(1)) * sigma + mu
        scale = np.where(np.random.rand(1) < self.scale_prob, scale, 1.)

        translate = self.trunc_normal_(low=-1., high=1., size=(2)) * self.translate
        translate = np.where(np.random.rand(1) < self.translate_prob, translate, 0.)

        bbox_center = bbox_center + bbox_wh * translate
        bbox_wh = bbox_wh * scale
        rot = (self.rotation * self.trunc_normal_(low=-1., high=1., size=(1))).item()

        # Get warping matrix
        warp_mat = self.get_warp_matrix(bbox_center, bbox_wh, rot)

        # Apply affine transform
        image = cv2.warpAffine(image, warp_mat, self.size, flags=cv2.INTER_LINEAR)
        image = Image.fromarray(image) # return as PIL

        # Compute keypoint. Note that this is only for one instance.
        # ``keypoint.shape`` should be (1, num_keypoints, 3)
        keypoint[..., :2] = cv2.transform(keypoint[..., :2], warp_mat)

        # Now, bbox is same with image size
        bbox_ = np.array([0, 0] + self.size[::-1]).astype('float32')
        bbox[0] = bbox_

        return image, label, mask, bbox, keypoint

    def __repr__(self):
        return self.__class__.__name__ + f"(scale={self.scale}, translate={self.translate}, rotation={self.rotation}, size={self.size})"

class Normalize:
    visualize = False

    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        image = F.normalize(image, mean=self.mean, std=self.std)
        return image, label, mask, bbox, keypoint

    def __repr__(self):
        return self.__class__.__name__ + "(mean={0}, std={1})".format(
            self.mean, self.std
        )


class ToTensor:
    visualize = False

    def __init__(self, pixel_range):
        self.pixel_range = pixel_range

    def __call__(self, image, label=None, mask=None, bbox=None, keypoint=None, dataset=None):
        image = F.to_tensor(image) * self.pixel_range
        if mask is not None:
            mask = torch.as_tensor(np.array(mask), dtype=torch.int64)
        if bbox is not None:
            bbox = torch.as_tensor(np.array(bbox), dtype=torch.float)

        return image, label, mask, bbox, keypoint

    def __repr__(self):
        return self.__class__.__name__ + "()"



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import math
import random
from typing import Any, Dict, List, Optional

import torch
import torchvision.transforms.functional as F
from torch.nn import functional as F_torch
from torchvision.transforms.functional import InterpolationMode


class Mixing:
    visualize = False

    def __init__(
        self,
        num_classes: int,
        cutmix: Optional[List],
        mixup: Optional[List],
        inplace: bool,
    ):
        self.mixup = bool(mixup)
        self.cutmix = bool(cutmix)
        self.num_classes = num_classes
        assert self.mixup or self.cutmix, "One of mixup or cutmix must be activated."

        self.transforms = []
        if self.mixup:
            assert len(mixup) == 2, "Mixup transform definition must be List of length 2."
            self.mixup_alpha, self.mixup_p = mixup
            self.transforms.append(RandomMixup(num_classes, self.mixup_alpha, self.mixup_p, inplace))

        if self.cutmix:
            assert len(cutmix) == 2, "Cutmix transform definition must be List of length 2."
            self.cutmix_alpha, self.cutmix_p = cutmix
            self.transforms.append(RandomCutmix(num_classes, self.cutmix_alpha, self.cutmix_p, inplace))

    def __call__(self, samples, targets):
        _mixup_transform = random.choice(self.transforms)
        return _mixup_transform(samples, targets)

    def __repr__(self) -> str:
        repr = "{}(num_classes={}, ".format(self.__class__.__name__, self.num_classes)
        if self.mixup:
            repr += "mixup_p={}, mixup_alpha={}, ".format(self.mixup_p, self.mixup_alpha)
        if self.cutmix:
            repr += "cutmix_p={}, alpha={}, ".format(self.cutmix_p, self.cutmix_alpha)
        repr += "inplace={})".format(self.inplace)
        return repr


class RandomMixup:
    """
    Based on the RandomMixup implementation of ml_cvnets.
    https://github.com/apple/ml-cvnets/blob/77717569ab4a852614dae01f010b32b820cb33bb/data/transforms/image_torch.py

    Given a batch of input images and labels, this class randomly applies the
    `MixUp transformation <https://arxiv.org/abs/1710.09412>`_

    Args:
        opts (argparse.Namespace): Arguments
        num_classes (int): Number of classes in the dataset
    """
    visualize = False

    def __init__(
        self,
        num_classes: int,
        alpha: float,
        p: float,
        inplace: bool,
    ):
        if not (num_classes > 0):
            raise ValueError("Please provide a valid positive value for the num_classes.")
        if not (alpha > 0):
            raise ValueError("Alpha param can't be zero.")
        if not (0.0 < p <= 1.0):
            raise ValueError("MixUp probability should be between 0 and 1, where 1 is inclusive")

        self.num_classes = num_classes
        self.alpha = alpha
        self.p = p
        self.inplace = inplace

    def _apply_mixup_transform(self, image_tensor, target_tensor):
        if image_tensor.ndim != 4:
            raise ValueError(f"Batch ndim should be 4. Got {image_tensor.ndim}")
        if target_tensor.ndim != 1:
            raise ValueError(f"Target ndim should be 1. Got {target_tensor.ndim}")
        if not image_tensor.is_floating_point():
            raise ValueError(f"Batch datatype should be a float tensor. Got {image_tensor.dtype}.")
        if target_tensor.dtype != torch.int64:
            raise ValueError(f"Target datatype should be torch.int64. Got {target_tensor.dtype}")

        if not self.inplace:
            image_tensor = image_tensor.clone()
            target_tensor = target_tensor.clone()

        if target_tensor.ndim == 1:
            target_tensor = F_torch.one_hot(
                target_tensor, num_classes=self.num_classes
            ).to(dtype=image_tensor.dtype)

        # It's faster to roll the batch by one instead of shuffling it to create image pairs
        batch_rolled = image_tensor.roll(1, 0)
        target_rolled = target_tensor.roll(1, 0)

        # Implemented as on mixup paper, page 3.
        lambda_param = float(
            torch._sample_dirichlet(torch.tensor([self.alpha, self.alpha]))[0]
        )
        batch_rolled.mul_(1.0 - lambda_param)
        image_tensor.mul_(lambda_param).add_(batch_rolled)

        target_rolled.mul_(1.0 - lambda_param)
        target_tensor.mul_(lambda_param).add_(target_rolled)
        return image_tensor, target_tensor

    def __call__(self, samples, targets):
        if torch.rand(1).item() >= self.p:
            return samples, targets

        mixup_samples, mixup_targets = self._apply_mixup_transform(
            image_tensor=samples, target_tensor=targets
        )

        return mixup_samples, mixup_targets

    def __repr__(self) -> str:
        return "{}(num_classes={}, p={}, alpha={}, inplace={})".format(
            self.__class__.__name__, self.num_classes, self.p, self.alpha, self.inplace
        )


class RandomCutmix:
    """
    Based on the RandomCutmix implementation of ml_cvnets.
    https://github.com/apple/ml-cvnets/blob/77717569ab4a852614dae01f010b32b820cb33bb/data/transforms/image_torch.py

    Given a batch of input images and labels, this class randomly applies the
    `CutMix transformation <https://arxiv.org/abs/1905.04899>`_

    Args:
        opts (argparse.Namespace): Arguments
        num_classes (int): Number of classes in the dataset
    """
    visualize = False

    def __init__(
        self,
        num_classes: int,
        alpha: float,
        p: float,
        inplace: bool,
    ):
        if not (num_classes > 0):
            raise ValueError("Please provide a valid positive value for the num_classes.")
        if not (alpha > 0):
            raise ValueError("Alpha param can't be zero.")
        if not (0.0 < p <= 1.0):
            raise ValueError("CutMix probability should be between 0 and 1, where 1 is inclusive")

        self.num_classes = num_classes
        self.alpha = alpha
        self.p = p
        self.inplace = inplace

    def _apply_cutmix_transform(self, image_tensor, target_tensor):
        if image_tensor.ndim != 4:
            raise ValueError(f"Batch ndim should be 4. Got {image_tensor.ndim}")
        if target_tensor.ndim != 1:
            raise ValueError(f"Target ndim should be 1. Got {target_tensor.ndim}")
        if not image_tensor.is_floating_point():
            raise ValueError(f"Batch dtype should be a float tensor. Got {image_tensor.dtype}.")
        if target_tensor.dtype != torch.int64:
            raise ValueError(f"Target dtype should be torch.int64. Got {target_tensor.dtype}")

        if not self.inplace:
            image_tensor = image_tensor.clone()
            target_tensor = target_tensor.clone()

        if target_tensor.ndim == 1:
            target_tensor = F_torch.one_hot(
                target_tensor, num_classes=self.num_classes
            ).to(dtype=image_tensor.dtype)

        # It's faster to roll the batch by one instead of shuffling it to create image pairs
        batch_rolled = image_tensor.roll(1, 0)
        target_rolled = target_tensor.roll(1, 0)

        # Implemented as on cutmix paper, page 12 (with minor corrections on typos).
        lambda_param = float(
            torch._sample_dirichlet(torch.tensor([self.alpha, self.alpha]))[0]
        )
        W, H = F.get_image_size(image_tensor)

        r_x = torch.randint(W, (1,))
        r_y = torch.randint(H, (1,))

        r = 0.5 * math.sqrt(1.0 - lambda_param)
        r_w_half = int(r * W)
        r_h_half = int(r * H)

        x1 = int(torch.clamp(r_x - r_w_half, min=0))
        y1 = int(torch.clamp(r_y - r_h_half, min=0))
        x2 = int(torch.clamp(r_x + r_w_half, max=W))
        y2 = int(torch.clamp(r_y + r_h_half, max=H))

        image_tensor[:, :, y1:y2, x1:x2] = batch_rolled[:, :, y1:y2, x1:x2]
        lambda_param = float(1.0 - (x2 - x1) * (y2 - y1) / (W * H))

        target_rolled.mul_(1.0 - lambda_param)
        target_tensor.mul_(lambda_param).add_(target_rolled)
        return image_tensor, target_tensor

    def __call__(self, samples, targets) -> Dict:
        if torch.rand(1).item() >= self.p:
            return samples, targets

        mixup_samples, mixup_targets = self._apply_cutmix_transform(
            image_tensor=samples, target_tensor=targets
        )

        return mixup_samples, mixup_targets

    def __repr__(self) -> str:
        return "{}(num_classes={}, p={}, alpha={}, inplace={})".format(
            self.__class__.__name__, self.num_classes, self.p, self.alpha, self.inplace
        )




# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
Based on the YOLO implementation of WongKinYiu
https://github.com/WongKinYiu/YOLO/blob/main/yolo/model/module.py
"""
import math
import torch
import torch.nn as nn

from omegaconf import DictConfig
from torch import Tensor
from torch.fx.proxy import Proxy
from typing import Dict, List, Optional, Tuple, Union
from ....op.custom import Anchor2Vec, ConvLayer
from ....utils import ModelOutput

def round_up(x: Union[int, Tensor], div: int = 1) -> Union[int, Tensor]:
    """
        Round up `x` to the biggest-nearest multiple of `div`
    """
    return x + (-x % div)


class Detection(nn.Module):
    """
        A single detection head.
    """
    def __init__(self,
                 in_channels: int,
                 hidden_channels: int,
                 num_classes: int,
                 act_type: Optional[str] = None,
                 reg_max: Optional[int] = 16,
                 use_group: bool = True,
                 prior_prob: Optional[float] = 1e-2
        ):
        super().__init__()
        groups = 4 if use_group else 1
        reg_channels = 4 * reg_max
        reg_hidden_channels = max(round_up(hidden_channels // 4, groups), reg_channels, reg_max)
        cls_hidden_channels = max(hidden_channels, min(num_classes * 2, 128))

        self.reg_convs = nn.Sequential(
            ConvLayer(in_channels, reg_hidden_channels, kernel_size=3, act_type=act_type),
            ConvLayer(reg_hidden_channels, reg_hidden_channels, kernel_size=3, groups=groups, act_type=act_type),
            nn.Conv2d(reg_hidden_channels, reg_channels, kernel_size=1, groups=groups)
        )

        self.cls_convs = nn.Sequential(
            ConvLayer(in_channels, cls_hidden_channels, kernel_size=3, act_type=act_type),
            ConvLayer(cls_hidden_channels, cls_hidden_channels, kernel_size=3, act_type=act_type),
            nn.Conv2d(cls_hidden_channels, num_classes, kernel_size=1)
        )

        self.anchor2vec = Anchor2Vec(reg_max=reg_max)

        # Initialize
        def init_bn(M):
            for m in M.modules():
                if isinstance(m, nn.BatchNorm2d):
                    m.eps = 1e-3
                    m.momentum = 0.03
        self.apply(init_bn)

        bias_reg = self.reg_convs[-1].bias.view(1, -1)
        bias_reg.data.fill_(-math.log((1 - prior_prob) / prior_prob))
        self.reg_convs[-1].bias = torch.nn.Parameter(bias_reg.view(-1), requires_grad=True)

        bias_cls = self.cls_convs[-1].bias.view(1, -1)
        bias_cls.data.fill_(-math.log((1 - prior_prob) / prior_prob))
        self.cls_convs[-1].bias = torch.nn.Parameter(bias_cls.view(-1), requires_grad=True)
    
    def forward(self, x: Union[Tensor, Proxy]) -> Tuple[Union[Tensor, Proxy]]:
        reg = self.reg_convs(x)
        cls_logits = self.cls_convs(x)
        output = torch.cat([reg, cls_logits], dim=1)
        return output


class YOLODetectionHead(nn.Module):
    def __init__(self,
                 num_classes: int,
                 intermediate_features_dim: List[int],
                 params: DictConfig):
        super().__init__()
        self._validate_params(params)
        self.num_classes = num_classes
        self.num_anchors = params.num_anchors
        self.hidden_dim = int(intermediate_features_dim[0])
        self.heads = self._build_heads(
            intermediate_features_dim,
            params.act_type,
            params.reg_max,
            params.use_group,
        )

        # if params.use_aux_loss:
        #     self.aux_heads = self._build_heads(
        #         intermediate_features_dim,
        #         params.act_type,
        #         params.reg_max,
        #         params.use_group,
        #     )
        # else:
        #     self.aux_heads = None

    def _validate_params(self, params: DictConfig) -> None:
        required_params = ['act_type', 'use_group', 'reg_max', 'num_anchors', 'use_aux_loss']
        for param in required_params:
            if not hasattr(params, param):
                raise ValueError(f"Missing required parameter: {param}")
    
    def _build_heads(
            self,
            intermediate_features_dim: List[int],
            act_type: str,
            reg_max: int,
            use_group: bool,
    ) -> nn.ModuleList:
        heads = nn.ModuleList()
        for feature_dim in intermediate_features_dim:
            head = Detection(
                int(feature_dim),
                self.hidden_dim,
                num_classes=self.num_classes,
                act_type=act_type,
                reg_max=reg_max,
                use_group=use_group,
            )
            heads.append(head)
        return heads

    def forward(self, x_in: Union[List[Tensor], Dict]) -> ModelOutput:
        if isinstance(x_in, Dict):
            assert self.aux_heads
            aux_in = x_in["aux_outputs"]
            x_in = x_in["outputs"]
        else:
            aux_in = None
        outputs = [head(x) for head, x in zip(self.heads, x_in)]
        # if self.training and self.aux_heads:
        #     aux_outputs = [head(x) for head, x in zip(self.aux_heads, aux_in)]
        #     outputs = {"outputs": outputs, "aux_outputs": aux_outputs}
        return ModelOutput(pred=outputs)

def yolo_detection_head(num_classes, intermediate_features_dim, conf_model_head, **kwargs):
    return YOLODetectionHead(num_classes=num_classes,
                             intermediate_features_dim=intermediate_features_dim,
                             params=conf_model_head.params)


# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import os
from abc import abstractmethod
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
from loguru import logger
from omegaconf import OmegaConf

from .utils import BackboneOutput, DetectionModelOutput, ModelOutput


class TaskModel(nn.Module):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone: bool = False) -> None:
        super(TaskModel, self).__init__()
        self.task = conf_model.task
        self.name = conf_model.name
        self.backbone_name = conf_model.architecture.backbone.name
        if neck:
            self.neck_name = conf_model.architecture.neck.name
        self.head_name = conf_model.architecture.head.name

        self.backbone = backbone
        if neck:
            self.neck = neck
        self.head = head

        if freeze_backbone:
            self._freeze_backbone()
            logger.info(f"Freeze! {self.backbone_name} is now freezed. Now only tuning with {self.head_name}.")
        self.__save_dtype = None

    def _freeze_backbone(self):
        for m in self.backbone.parameters():
            m.requires_grad = False

    @property
    def head_list(self):
        return ('head')

    @property
    def device(self):
        return next(self.parameters()).device

    @property
    def save_dtype(self):
        return self.__save_dtype

    @save_dtype.setter
    def save_dtype(self, dtype):
        self.__save_dtype = dtype

    def _get_name(self):
        if hasattr(self, 'neck'):
            return f"{self.__class__.__name__}[task={self.task}, backbone={self.backbone_name}, neck={self.neck_name}, head={self.head_name}]"
        else:
            return f"{self.__class__.__name__}[task={self.task}, backbone={self.backbone_name}, head={self.head_name}]"

    @abstractmethod
    def forward(self, x):
        raise NotImplementedError

    def deploy(self, ):
        self.eval()
        for m in self.modules():
            if hasattr(m, 'convert_to_deploy'):
                m.convert_to_deploy()
        return self


class ClassificationModel(TaskModel):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone=False) -> None:
        super().__init__(conf_model, backbone, neck, head, freeze_backbone)

    def forward(self, x):
        features: BackboneOutput = self.backbone(x)
        out: ModelOutput = self.head(features['last_feature'])
        return out


class SegmentationModel(TaskModel):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone=False) -> None:
        super().__init__(conf_model, backbone, neck, head, freeze_backbone)

    def forward(self, x):
        features: BackboneOutput = self.backbone(x)
        if hasattr(self, 'neck'):
            features: BackboneOutput = self.neck(features['intermediate_features'])
        out: ModelOutput = self.head(features['intermediate_features'])
        return out


class DetectionModel(TaskModel):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone=False) -> None:
        super().__init__(conf_model, backbone, neck, head, freeze_backbone)

    def forward(self, x):
        features: BackboneOutput = self.backbone(x)
        if hasattr(self, 'neck'):
            features: BackboneOutput = self.neck(features['intermediate_features'])
        out: DetectionModelOutput = self.head(features['intermediate_features'])
        return out


class PoseEstimationModel(TaskModel):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone=False) -> None:
        super().__init__(conf_model, backbone, neck, head, freeze_backbone)

    def forward(self, x):
        features: BackboneOutput = self.backbone(x)
        if hasattr(self, 'neck'):
            features: BackboneOutput = self.neck(features['intermediate_features'])
        out: DetectionModelOutput = self.head(features['intermediate_features'])
        return out


class ONNXModel:
    '''
        ONNX Model wrapper class for inferencing.
    '''
    def __init__(self, model_conf) -> None:
        import onnxruntime as ort
        self.name = model_conf.name + '_onnx'
        self.onnx_path = model_conf.checkpoint.path
        providers = [
            ('CUDAExecutionProvider', {
                'device_id': 0,
            }),
            'CPUExecutionProvider',
        ]
        self.inference_session = ort.InferenceSession(model_conf.checkpoint.path, providers=providers)

    def _get_name(self):
        return f"{self.__class__.__name__}[model={self.name}]"

    def __call__(self, x, targets=None):
        device = x.device
        x = x.detach().cpu().numpy()
        out = self.inference_session.run(None, {self.inference_session.get_inputs()[0].name: x})
        out = [torch.tensor(o).to(device) for o in out]

        if len(out) == 1:
            out = out[0]

        return ModelOutput(pred=out)

    def eval(self):
        pass # Do nothing

    def set_provider(self, device):
        if device.type == 'cuda':
            self.inference_session.set_providers(['CUDAExecutionProvider'])
        else:
            self.inference_session.set_providers(['CPUExecutionProvider'])


class TFLiteModel:
    """
    TensorFlow Lite (tflite) wrapper class for inferencing.
    """

    NUM_THREADS = 4

    def __init__(self, model_conf) -> None:
        self.tflite = self._import_tflite()
        self.task = model_conf.task
        assert self.task == 'detection', f"Task {self.task} is not yet supported in this TensorFlow Lite (tflite) model inference."
        self.name = model_conf.name + '_tflite'
        self.tflite_path = model_conf.checkpoint.path
        self.interpreter = self.tflite.Interpreter(model_path=self.tflite_path, num_threads=self.NUM_THREADS)
        self.interpreter.allocate_tensors()
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()
        self.input_dtype = self.input_details[0]['dtype']
        self.input_shape = tuple(self.input_details[0]['shape'])

        self.output_dtype = self.output_details[0]['dtype']
        self.quantized_input = self.input_dtype in [np.int8, np.uint8]
        self.quantized_output = self.output_dtype in [np.int8, np.uint8]
        if self.quantized_input:
            self.input_scale, self.input_zero_point = self.input_details[0]['quantization']

    def _import_tflite(self):
        try:
            import tflite_runtime.interpreter as tflite
        except ImportError:
            try:
                import tensorflow.lite as tflite
            except ImportError as e:
                raise ImportError("Failed to import tensorflow lite. Please install tflite_runtime or tensorflow") from e
        return tflite

    def get_name(self) -> str:
        """Get the name of the model."""
        return f"{self.__class__.__name__}[model={self.name}]"

    def __call__(self, x: Union[np.ndarray, torch.Tensor], targets=None):
        """
        Perform inference on the input tensor.

        Args:
            x (Union[np.ndarray, torch.Tensor]): Input tensor
            targets: Not used in this implementation

        Returns:
            ModelOutput: Output of the model
        """
        try:
            device = x.device if hasattr(x, 'device') else 'cpu'
            x = self._prepare_input(x)

            if self.quantized_input:
                x = x / self.input_scale + self.input_zero_point
                x = x.astype(self.input_dtype)

            assert x.shape == self.input_shape, f"Your input shape {x.shape} does not match with the expected input shape {self.input_shape}"
            self.interpreter.set_tensor(self.input_details[0]['index'], x)
            self.interpreter.invoke()

            output = self._process_output(device)
            return ModelOutput(pred=output)
        except Exception as e:
            raise RuntimeError(f"Error during inference: {str(e)}") from e

    def _process_output(self, device: str) -> List[torch.Tensor]:
        """Process the output of the interpreter."""
        output = []
        for details in self.output_details:
            o = self.interpreter.get_tensor(details["index"])
            if self.quantized_output:
                output_quantization_params = details['quantization']
                o = (o.astype(np.float32) - output_quantization_params[1]) * output_quantization_params[0]
            output.append(torch.tensor(np.transpose(o, (0, 3, 1, 2))).to(device))

        if len(output) > 1:
            output.sort(key=lambda x: sum(x.shape), reverse=True)
        return output

    def eval(self):
        """Set the model to evaluation mode."""
        pass  # Do nothing for TFLite model

    def _prepare_input(self, x: Union[np.ndarray, torch.Tensor]) -> np.ndarray:
        """
        Prepare input tensor for inference.

        Args:
            x (Union[np.ndarray, torch.Tensor]): Input tensor

        Returns:
            np.ndarray: Prepared input tensor
        """
        if isinstance(x, torch.Tensor):
            x = x.detach().cpu().numpy()
        elif not isinstance(x, np.ndarray):
            raise TypeError(f"Unsupported input type: {type(x)}")

        if x.shape[1] == 3:
            x = np.transpose(x, (0, 2, 3, 1))
        return x



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

# Modified from official PIDNet implementation:
# https://github.com/XuJiacong/PIDNet

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..op.custom import Pool2d

BatchNorm2d = nn.BatchNorm2d
bn_mom = 0.1
algc = False


# class BasicBlock(nn.Module):
#     expansion = 1

#     def __init__(self, inplanes, planes, stride=1, downsample=None, no_relu=False):
#         super(BasicBlock, self).__init__()
#         self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=3, stride=stride,
#                                padding=1, bias=False)
#         self.bn1 = BatchNorm2d(planes, momentum=bn_mom)
#         self.relu = nn.ReLU(inplace=True)
#         self.conv2 = nn.Conv2d(planes, planes, kernel_size=3,
#                                padding=1, bias=False)
#         self.bn2 = BatchNorm2d(planes, momentum=bn_mom)
#         self.downsample = downsample
#         self.stride = stride
#         self.no_relu = no_relu

#     def forward(self, x):
#         residual = x

#         out = self.conv1(x)
#         out = self.bn1(out)
#         out = self.relu(out)

#         out = self.conv2(out)
#         out = self.bn2(out)

#         if self.downsample is not None:
#             residual = self.downsample(x)

#         out += residual

#         if self.no_relu:
#             return out
#         else:
#             return self.relu(out)


# class Bottleneck(nn.Module):
#     expansion = 2

#     def __init__(self, inplanes, planes, stride=1, downsample=None, no_relu=True):
#         super(Bottleneck, self).__init__()
#         self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
#         self.bn1 = BatchNorm2d(planes, momentum=bn_mom)
#         self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
#                                padding=1, bias=False)
#         self.bn2 = BatchNorm2d(planes, momentum=bn_mom)
#         self.conv3 = nn.Conv2d(planes, planes * self.expansion, kernel_size=1,
#                                bias=False)
#         self.bn3 = BatchNorm2d(planes * self.expansion, momentum=bn_mom)
#         self.relu = nn.ReLU(inplace=True)
#         self.downsample = downsample
#         self.stride = stride
#         self.no_relu = no_relu

#     def forward(self, x):
#         residual = x

#         out = self.conv1(x)
#         out = self.bn1(out)
#         out = self.relu(out)

#         out = self.conv2(out)
#         out = self.bn2(out)
#         out = self.relu(out)

#         out = self.conv3(out)
#         out = self.bn3(out)

#         if self.downsample is not None:
#             residual = self.downsample(x)

#         out += residual
#         if self.no_relu:
#             return out
#         else:
#             return self.relu(out)


class segmenthead(nn.Module):

    def __init__(self, inplanes, interplanes, outplanes, scale_factor=None):
        super(segmenthead, self).__init__()
        self.bn1 = BatchNorm2d(inplanes, momentum=bn_mom)
        self.conv1 = nn.Conv2d(inplanes, interplanes, kernel_size=3, padding=1, bias=False)
        self.bn2 = BatchNorm2d(interplanes, momentum=bn_mom)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(interplanes, outplanes, kernel_size=1, padding=0, bias=True)
        self.scale_factor = scale_factor

    def forward(self, x):

        x = self.conv1(self.relu(self.bn1(x)))
        out = self.conv2(self.relu(self.bn2(x)))

        # if self.scale_factor is not None:
        #     height = x.shape[-2] * self.scale_factor
        #     width = x.shape[-1] * self.scale_factor
        #     out = F.interpolate(out,
        #                         size=[height, width],
        #                         mode='bilinear', align_corners=algc)

        return out


class DAPPM(nn.Module):
    def __init__(self, inplanes, branch_planes, outplanes, BatchNorm=nn.BatchNorm2d):
        super(DAPPM, self).__init__()
        bn_mom = 0.1
        self.scale1 = nn.Sequential(Pool2d(method='avg', kernel_size=5, stride=2, padding=2),
                                    BatchNorm(inplanes, momentum=bn_mom),
                                    nn.ReLU(inplace=True),
                                    nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
                                    )
        self.scale2 = nn.Sequential(Pool2d(method='avg', kernel_size=9, stride=4, padding=4),
                                    BatchNorm(inplanes, momentum=bn_mom),
                                    nn.ReLU(inplace=True),
                                    nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
                                    )
        self.scale3 = nn.Sequential(Pool2d(method='avg', kernel_size=17, stride=8, padding=8),
                                    BatchNorm(inplanes, momentum=bn_mom),
                                    nn.ReLU(inplace=True),
                                    nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
                                    )
        self.scale4 = nn.Sequential(nn.AdaptiveAvgPool2d((1, 1)),
                                    BatchNorm(inplanes, momentum=bn_mom),
                                    nn.ReLU(inplace=True),
                                    nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
                                    )
        self.scale0 = nn.Sequential(
            BatchNorm(inplanes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
        )
        self.process1 = nn.Sequential(
            BatchNorm(branch_planes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(branch_planes, branch_planes, kernel_size=3, padding=1, bias=False),
        )
        self.process2 = nn.Sequential(
            BatchNorm(branch_planes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(branch_planes, branch_planes, kernel_size=3, padding=1, bias=False),
        )
        self.process3 = nn.Sequential(
            BatchNorm(branch_planes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(branch_planes, branch_planes, kernel_size=3, padding=1, bias=False),
        )
        self.process4 = nn.Sequential(
            BatchNorm(branch_planes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(branch_planes, branch_planes, kernel_size=3, padding=1, bias=False),
        )
        self.compression = nn.Sequential(
            BatchNorm(branch_planes * 5, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(branch_planes * 5, outplanes, kernel_size=1, bias=False),
        )
        self.shortcut = nn.Sequential(
            BatchNorm(inplanes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(inplanes, outplanes, kernel_size=1, bias=False),
        )

    def forward(self, x):
        x.shape[-1]
        x.shape[-2]
        x_list = []

        x_list.append(self.scale0(x))
        x_list.append(self.process1((F.interpolate(self.scale1(x),
                                                   size=x_list[0].shape[-2:],
                                                   mode='bilinear', align_corners=algc)+x_list[0])))
        x_list.append(self.process2((F.interpolate(self.scale2(x),
                                                   size=x_list[1].shape[-2:],
                                                   mode='bilinear', align_corners=algc)+x_list[1])))
        x_list.append(self.process3((F.interpolate(self.scale3(x),
                                                   size=x_list[2].shape[-2:],
                                                   mode='bilinear', align_corners=algc)+x_list[2])))
        x_list.append(self.process4((F.interpolate(self.scale4(x),
                                                   size=x_list[3].shape[-2:],
                                                   mode='bilinear', align_corners=algc)+x_list[3])))

        out = self.compression(torch.cat(x_list, 1)) + self.shortcut(x)
        return out


class PAPPM(nn.Module):
    def __init__(self, inplanes, branch_planes, outplanes, BatchNorm=nn.BatchNorm2d):
        super(PAPPM, self).__init__()
        bn_mom = 0.1
        self.scale1 = nn.Sequential(Pool2d(method='avg', kernel_size=5, stride=2, padding=2),
                                    BatchNorm(inplanes, momentum=bn_mom),
                                    nn.ReLU(inplace=True),
                                    nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
                                    )
        self.scale2 = nn.Sequential(Pool2d(method='avg', kernel_size=9, stride=4, padding=4),
                                    BatchNorm(inplanes, momentum=bn_mom),
                                    nn.ReLU(inplace=True),
                                    nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
                                    )
        self.scale3 = nn.Sequential(Pool2d(method='avg', kernel_size=17, stride=8, padding=8),
                                    BatchNorm(inplanes, momentum=bn_mom),
                                    nn.ReLU(inplace=True),
                                    nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
                                    )
        self.scale4 = nn.Sequential(nn.AdaptiveAvgPool2d((1, 1)),
                                    BatchNorm(inplanes, momentum=bn_mom),
                                    nn.ReLU(inplace=True),
                                    nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
                                    )

        self.scale0 = nn.Sequential(
            BatchNorm(inplanes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
        )

        self.scale_process = nn.Sequential(
            BatchNorm(branch_planes*4, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(branch_planes*4, branch_planes*4, kernel_size=3, padding=1, groups=4, bias=False),
        )

        self.compression = nn.Sequential(
            BatchNorm(branch_planes * 5, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(branch_planes * 5, outplanes, kernel_size=1, bias=False),
        )

        self.shortcut = nn.Sequential(
            BatchNorm(inplanes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(inplanes, outplanes, kernel_size=1, bias=False),
        )

    def forward(self, x):
        x.shape[-1]
        x.shape[-2]
        scale_list = []

        x_ = self.scale0(x)
        scale_list.append(F.interpolate(self.scale1(x), size=x_.shape[-2:],
                                        mode='bilinear', align_corners=algc)+x_)
        scale_list.append(F.interpolate(self.scale2(x), size=x_.shape[-2:],
                                        mode='bilinear', align_corners=algc)+x_)
        scale_list.append(F.interpolate(self.scale3(x), size=x_.shape[-2:],
                                        mode='bilinear', align_corners=algc)+x_)
        scale_list.append(F.interpolate(self.scale4(x), size=x_.shape[-2:],
                                        mode='bilinear', align_corners=algc)+x_)

        scale_out = self.scale_process(torch.cat(scale_list, 1))

        out = self.compression(torch.cat([x_, scale_out], 1)) + self.shortcut(x)
        return out


class PagFM(nn.Module):
    def __init__(self, in_channels, mid_channels, after_relu=False, with_channel=False, BatchNorm=nn.BatchNorm2d):
        super(PagFM, self).__init__()
        self.with_channel = with_channel
        self.after_relu = after_relu
        self.f_x = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels,
                      kernel_size=1, bias=False),
            BatchNorm(mid_channels)
        )
        self.f_y = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels,
                      kernel_size=1, bias=False),
            BatchNorm(mid_channels)
        )
        if with_channel:
            self.up = nn.Sequential(
                nn.Conv2d(mid_channels, in_channels,
                          kernel_size=1, bias=False),
                BatchNorm(in_channels)
            )
        if after_relu:
            self.relu = nn.ReLU(inplace=True)

    def forward(self, x, y):

        if self.after_relu:
            y = self.relu(y)
            x = self.relu(x)

        y_q = self.f_y(y)
        y_q = F.interpolate(y_q, size=x.shape[-2:],
                            mode='bilinear', align_corners=False)
        x_k = self.f_x(x)

        if self.with_channel:
            sim_map = torch.sigmoid(self.up(x_k * y_q))
        else:
            sim_map = torch.sigmoid(torch.sum(x_k * y_q, dim=1).unsqueeze(1))

        y = F.interpolate(y, size=x.shape[-2:],
                          mode='bilinear', align_corners=False)
        x = (1-sim_map)*x + sim_map*y

        return x


class Light_Bag(nn.Module):
    def __init__(self, in_channels, out_channels, BatchNorm=nn.BatchNorm2d):
        super(Light_Bag, self).__init__()
        self.conv_p = nn.Sequential(
            nn.Conv2d(in_channels, out_channels,
                      kernel_size=1, bias=False),
            BatchNorm(out_channels)
        )
        self.conv_i = nn.Sequential(
            nn.Conv2d(in_channels, out_channels,
                      kernel_size=1, bias=False),
            BatchNorm(out_channels)
        )

    def forward(self, p, i, d):
        edge_att = torch.sigmoid(d)

        p_add = self.conv_p((1-edge_att)*i + p)
        i_add = self.conv_i(i + edge_att*p)

        return p_add + i_add


class DDFMv2(nn.Module):
    def __init__(self, in_channels, out_channels, BatchNorm=nn.BatchNorm2d):
        super(DDFMv2, self).__init__()
        self.conv_p = nn.Sequential(
            BatchNorm(in_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels, out_channels,
                      kernel_size=1, bias=False),
            BatchNorm(out_channels)
        )
        self.conv_i = nn.Sequential(
            BatchNorm(in_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels, out_channels,
                      kernel_size=1, bias=False),
            BatchNorm(out_channels)
        )

    def forward(self, p, i, d):
        edge_att = torch.sigmoid(d)

        p_add = self.conv_p((1-edge_att)*i + p)
        i_add = self.conv_i(i + edge_att*p)

        return p_add + i_add


class Bag(nn.Module):
    def __init__(self, in_channels, out_channels, BatchNorm=nn.BatchNorm2d):
        super(Bag, self).__init__()

        self.conv = nn.Sequential(
            BatchNorm(in_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels, out_channels,
                      kernel_size=3, padding=1, bias=False)
        )

    def forward(self, p, i, d):
        edge_att = torch.sigmoid(d)
        return self.conv(edge_att*p + (1-edge_att)*i)


if __name__ == '__main__':

    x = torch.rand(4, 64, 32, 64).cuda()
    y = torch.rand(4, 64, 32, 64).cuda()
    z = torch.rand(4, 64, 32, 64).cuda()
    net = PagFM(64, 16, with_channel=True).cuda()

    out = net(x, y)
